"""
*******************************************************
 * Copyright (C) 2017 MindsDB Inc. <copyright@mindsdb.com>
 *
 * This file is part of MindsDB Server.
 *
 * MindsDB Server can not be copied and/or distributed without the express
 * permission of MindsDB Inc
 *******************************************************
"""
import copy
import re
import hashlib
import datetime as dt
from collections import defaultdict

import dateinfer
import pandas as pd
import numpy as np

from mindsdb_sql import parse_sql
from mindsdb_sql.parser.ast import (
    BinaryOperation,
    UnaryOperation,
    Identifier,
    Constant,
    Select,
    Union,
    Join,
    Star,
    Insert,
    Update,
    Delete,
    Latest,
    BetweenOperation,
    Parameter,
    Tuple,
)
from mindsdb_sql.planner.step_result import Result
from mindsdb_sql.planner.steps import (
    ApplyTimeseriesPredictorStep,
    ApplyPredictorRowStep,
    GetPredictorColumns,
    FetchDataframeStep,
    ApplyPredictorStep,
    GetTableColumns,
    LimitOffsetStep,
    MapReduceStep,
    MultipleSteps,
    ProjectStep,
    SaveToTable,
    InsertToTable,
    UpdateToTable,
    FilterStep,
    UnionStep,
    JoinStep,
    GroupByStep,
    SubSelectStep,
    DeleteStep,
    DataStep,
)

from mindsdb_sql.exceptions import PlanningException
from mindsdb_sql.render.sqlalchemy_render import SqlalchemyRender
from mindsdb_sql.planner import query_planner
from mindsdb_sql.planner.utils import query_traversal

from mindsdb.api.mysql.mysql_proxy.utilities.sql import query_df, query_df_with_type_infer_fallback
from mindsdb.interfaces.model.functions import get_model_record
from mindsdb.api.mysql.mysql_proxy.utilities import (
    ErKeyColumnDoesNotExist,
    ErNotSupportedYet,
    SqlApiUnknownError,
    ErLogicError,
    ErSqlWrongArguments
)
from mindsdb.interfaces.query_context.context_controller import query_context_controller
from mindsdb.utilities.cache import get_cache, json_checksum
import mindsdb.utilities.profiler as profiler
from mindsdb.utilities.fs import create_process_mark, delete_process_mark


superset_subquery = re.compile(r'from[\s\n]*(\(.*\))[\s\n]*as[\s\n]*virtual_table', flags=re.IGNORECASE | re.MULTILINE | re.S)


def get_preditor_alias(step, mindsdb_database):
    predictor_name = '.'.join(step.predictor.parts)
    predictor_alias = '.'.join(step.predictor.alias.parts) if step.predictor.alias is not None else predictor_name
    return (mindsdb_database, predictor_name, predictor_alias)


def get_table_alias(table_obj, default_db_name):
    # (database, table, alias)
    if isinstance(table_obj, Identifier):
        if len(table_obj.parts) == 1:
            name = (default_db_name, table_obj.parts[0])
        else:
            name = (table_obj.parts[0], table_obj.parts[-1])
    elif isinstance(table_obj, Select):
        # it is subquery
        if table_obj.alias is None:
            name = 't'
        else:
            name = table_obj.alias.parts[0]
        name = (default_db_name, name)
    elif isinstance(table_obj, Join):
        # get from first table
        return get_table_alias(table_obj.left, default_db_name)
    else:
        # unknown yet object
        return default_db_name, 't', 't'

    if table_obj.alias is not None:
        name = name + ('.'.join(table_obj.alias.parts),)
    else:
        name = name + (name[1],)
    return name


def get_all_tables(stmt):
    if isinstance(stmt, Union):
        left = get_all_tables(stmt.left)
        right = get_all_tables(stmt.right)
        return left + right

    if isinstance(stmt, Select):
        from_stmt = stmt.from_table
    elif isinstance(stmt, (Identifier, Join)):
        from_stmt = stmt
    elif isinstance(stmt, Insert):
        from_stmt = stmt.table
    elif isinstance(stmt, Delete):
        from_stmt = stmt.table
    else:
        # raise SqlApiException(f'Unknown type of identifier: {stmt}')
        return []

    result = []
    if isinstance(from_stmt, Identifier):
        result.append(from_stmt.parts[-1])
    elif isinstance(from_stmt, Join):
        result.extend(get_all_tables(from_stmt.left))
        result.extend(get_all_tables(from_stmt.right))
    return result


def markQueryVar(where):
    if isinstance(where, BinaryOperation):
        markQueryVar(where.args[0])
        markQueryVar(where.args[1])
    elif isinstance(where, UnaryOperation):
        markQueryVar(where.args[0])
    elif isinstance(where, Constant):
        if str(where.value).startswith('$var['):
            where.is_var = True
            where.var_name = where.value


def unmarkQueryVar(where):
    if isinstance(where, BinaryOperation):
        unmarkQueryVar(where.args[0])
        unmarkQueryVar(where.args[1])
    elif isinstance(where, UnaryOperation):
        unmarkQueryVar(where.args[0])
    elif isinstance(where, Constant):
        if hasattr(where, 'is_var') and where.is_var is True:
            where.value = where.var_name


def replaceQueryVar(where, var_value, var_name):
    if isinstance(where, BinaryOperation):
        replaceQueryVar(where.args[0], var_value, var_name)
        replaceQueryVar(where.args[1], var_value, var_name)
    elif isinstance(where, UnaryOperation):
        replaceQueryVar(where.args[0], var_value, var_name)
    elif isinstance(where, Constant):
        if hasattr(where, 'is_var') and where.is_var is True and where.value == f'$var[{var_name}]':
            where.value = var_value


def join_query_data(target, source):
    if len(target.columns) == 0:
        target = source
    else:
        target.add_records(source.get_records())
    return target


def is_empty_prediction_row(predictor_value):
    "Define empty rows in predictor after JOIN"
    for key in predictor_value:
        if predictor_value[key] is not None and pd.notna(predictor_value[key]):
            return False
    return True


class Column:
    def __init__(self, name=None, alias=None,
                 table_name=None, table_alias=None,
                 type=None, database=None, flags=None,
                 charset=None):
        if alias is None:
            alias = name
        if table_alias is None:
            table_alias = table_name
        self.name = name
        self.alias = alias
        self.table_name = table_name
        self.table_alias = table_alias
        self.type = type
        self.database = database
        self.flags = flags
        self.charset = charset

    def get_hash_name(self, prefix):
        table_name = self.table_name if self.table_alias is None else self.table_alias
        name = self.name if self.alias is None else self.alias

        name = f'{prefix}_{table_name}_{name}'
        return name

    def __repr__(self):
        return f'{self.__class__.__name__}({self.__dict__})'


class ResultSet:
    def __init__(self, length=0):
        self._columns = []
        # records is list of lists with the same length as columns
        self._records = []
        for i in range(length):
            self._records.append([])

        self.is_prediction = False

    def __repr__(self):
        col_names = ', '.join([col.name for col in self._columns])
        data = '\n'.join([str(rec) for rec in self._records[:20]])

        if len(self._records) > 20:
            data += '\n...'

        return f'{self.__class__.__name__}({self.length()} rows, cols: {col_names})\n {data}'

    # --- converters ---

    def from_df(self, df, database, table_name, table_alias=None):

        resp_dict = df.to_dict(orient='split')

        self._records = resp_dict['data']

        for col in resp_dict['columns']:
            self._columns.append(Column(
                name=col,
                table_name=table_name,
                table_alias=table_alias,
                database=database,
                type=df.dtypes[col]
            ))
        return self

    def from_df_cols(self, df, col_names, strict=True):
        # find column by alias
        alias_idx = {}
        for col in col_names.values():
            if col.alias is not None:
                alias_idx[col.alias] = col

        resp_dict = df.to_dict(orient='split')

        self._records = resp_dict['data']

        for col in resp_dict['columns']:
            if col in col_names or strict:
                column = col_names[col]
            elif col in alias_idx:
                column = alias_idx[col]
            else:
                column = Column(col)
            self._columns.append(column)
        return self

    def to_df(self):
        columns = self.get_column_names()
        return pd.DataFrame(self._records, columns=columns)

    def to_df_cols(self, prefix=''):
        # returns dataframe and dict of columns
        #   can be restored to ResultSet by from_df_cols method

        columns = []
        col_names = {}
        for col in self._columns:
            name = col.get_hash_name(prefix)
            columns.append(name)
            col_names[name] = col

        return pd.DataFrame(self._records, columns=columns), col_names

    # --- tables ---

    def get_tables(self):
        tables_idx = []
        tables = []
        cols = ['database', 'table_name', 'table_alias']
        for col in self._columns:
            table = (col.database, col.table_name, col.table_alias)
            if table not in tables_idx:
                tables_idx.append(table)
                tables.append(dict(zip(cols, table)))
        return tables

    # --- columns ---

    def _locate_column(self, col):
        col_idx = None
        for i, col0 in enumerate(self._columns):
            if col0 is col:
                col_idx = i
                break
        if col_idx is None:
            raise ErSqlWrongArguments(f'Column is not found: {col}')
        return col_idx

    def add_column(self, col, values=None):
        self._columns.append(col)

        if values is None:
            values = []
        # update records
        if len(self._records) > 0:
            for rec in self._records:
                if len(values) > 0:
                    value = values.pop(0)
                else:
                    value = None
                rec.append(value)

    def del_column(self, col):
        idx = self._locate_column(col)
        self._columns.pop(idx)
        for row in self._records:
            row.pop(idx)

    @property
    def columns(self):
        return self._columns

    def get_column_names(self):
        columns = [
            col.name if col.alias is None else col.alias
            for col in self._columns
        ]
        return columns

    def find_columns(self, alias=None, table_alias=None):
        col_list = []
        for col in self.columns:
            if alias is not None and col.alias.lower() != alias.lower():
                continue
            if table_alias is not None and col.table_alias.lower() != table_alias.lower():
                continue
            col_list.append(col)

        return col_list

    def copy_column_to(self, col, result_set2):
        # copy with values
        idx = self._locate_column(col)

        values = [row[idx] for row in self._records]

        col2 = copy.deepcopy(col)

        result_set2.add_column(col2, values)
        return col2

    # --- records ---

    def add_records(self, data):
        names = self.get_column_names()
        for rec in data:
            # if len(rec) != len(self._columns):
            #     raise ErSqlWrongArguments(f'Record length mismatch columns length: {len(rec)} != {len(self._columns)}')

            record = [
                rec[name]
                for name in names
            ]
            self._records.append(record)

    def get_records_raw(self):
        return self._records

    def add_record_raw(self, rec):
        if len(rec) != len(self._columns):
            raise ErSqlWrongArguments(f'Record length mismatch columns length: {len(rec)} != {len(self.columns)}')
        self._records.append(rec)

    @property
    def records(self):
        return self.get_records()

    def get_records(self):
        # get records as dicts.
        # !!! Attention: !!!
        # if resultSet contents duplicate column name: only one of them will be in output
        names = self.get_column_names()
        records = []
        for row in self._records:
            records.append(dict(zip(names, row)))
        return records

    # def clear_records(self):
    #     self._records = []

    def length(self):
        return len(self._records)


class SQLQuery():
    def __init__(self, sql, session, execute=True):
        self.session = session
        self.database = None if session.database == '' else session.database.lower()
        self.datahub = session.datahub
        self.outer_query = None
        self.row_id = 0
        self.columns_list = None
        self.model_types = {}

        self.mindsdb_database_name = 'mindsdb'

        if isinstance(sql, str):
            # region workaround for subqueries in superset
            if 'as virtual_table' in sql.lower():
                subquery = re.findall(superset_subquery, sql)
                if isinstance(subquery, list) and len(subquery) == 1:
                    subquery = subquery[0]
                    self.outer_query = sql.replace(subquery, 'dataframe')
                    sql = subquery.strip('()')
            # endregion
            self.query = parse_sql(sql, dialect='mindsdb')
            self.query_str = sql
        else:
            self.query = sql
            renderer = SqlalchemyRender('mysql')
            try:
                self.query_str = renderer.get_string(self.query, with_failback=True)
            except Exception:
                self.query_str = str(self.query)

        self.planner = None
        self.parameters = []
        self.fetched_data = None
        # self._process_query(sql)
        self.create_planner()
        if execute:
            self.prepare_query(prepare=False)
            self.execute_query()

    @profiler.profile()
    def create_planner(self):
        databases = self.session.database_controller.get_list()

        predictor_metadata = []

        query_tables = []

        def get_all_query_tables(node, is_table, **kwargs):
            if is_table and isinstance(node, Identifier):
                table_name = node.parts[-1]
                table_version = None
                project_name = self.session.database
                if table_name.isdigit():
                    # is predictor version
                    table_version = int(table_name)
                    table_name = node.parts[-2]
                if table_name != node.parts[0]:
                    project_name = node.parts[0]
                query_tables.append((table_name, table_version, project_name))

        query_traversal(self.query, get_all_query_tables)

        for table_name, table_version, project_name in query_tables:
            args = {
                'name': table_name,
                'project_name': project_name
            }
            if table_version is not None:
                args['active'] = None
                args['version'] = table_version

            model_record = get_model_record(**args)
            if model_record is None:
                continue

            if isinstance(model_record.data, dict) is False or 'error' in model_record.data:
                continue

            ts_settings = model_record.learn_args.get('timeseries_settings', {})
            predictor = {
                'name': table_name,
                'integration_name': project_name,   # integration_name,
                'timeseries': False,
                'id': model_record.id
            }
            if ts_settings.get('is_timeseries') is True:
                window = ts_settings.get('window')
                order_by = ts_settings.get('order_by')
                if isinstance(order_by, list):
                    order_by = order_by[0]
                group_by = ts_settings.get('group_by')
                if isinstance(group_by, list) is False and group_by is not None:
                    group_by = [group_by]
                predictor.update({
                    'timeseries': True,
                    'window': window,
                    'horizon': ts_settings.get('horizon'),
                    'order_by_column': order_by,
                    'group_by_columns': group_by
                })
            predictor_metadata.append(predictor)

            self.model_types.update(model_record.data.get('dtypes', {}))

        database = None if self.session.database == '' else self.session.database.lower()

        self.predictor_metadata = predictor_metadata
        self.planner = query_planner.QueryPlanner(
            self.query,
            integrations=databases,
            predictor_metadata=predictor_metadata,
            default_namespace=database,
        )

    def fetch(self, view='list'):
        data = self.fetched_data

        if view == 'dataframe':
            result = data.to_df()
        else:
            result = data.get_records_raw()

        return {
            'success': True,
            'result': result
        }

    def _fetch_dataframe_step(self, step, steps_data):
        dn = self.datahub.get(step.integration)
        query = step.query

        if dn is None:
            raise SqlApiUnknownError(f'Unknown integration name: {step.integration}')

        if query is None:
            table_alias = (self.database, 'result', 'result')

            # fetch raw_query
            data, columns_info = dn.query(
                native_query=step.raw_query,
                session=self.session
            )
        else:
            table_alias = get_table_alias(step.query.from_table, self.database)
            # TODO for information_schema we have 'database' = 'mindsdb'

            # fill params
            def fill_params(node, **kwargs):
                if isinstance(node, Parameter):
                    rs = steps_data[node.value.step_num]
                    items = [Constant(i[0]) for i in rs.get_records_raw()]
                    return Tuple(items)

            query_traversal(query, fill_params)

            query, context_callback = query_context_controller.handle_db_context_vars(query, dn, self.session)

            data, columns_info = dn.query(
                query=query,
                session=self.session
            )

            if context_callback:
                context_callback(data, columns_info)

        result = ResultSet()
        for column in columns_info:
            result.add_column(Column(
                name=column['name'],
                type=column.get('type'),
                table_name=table_alias[1],
                table_alias=table_alias[2],
                database=table_alias[0]
            ))
        result.add_records(data)

        return result

    def _multiple_steps(self, steps, steps_data):
        data = ResultSet()
        for substep in steps:
            sub_data = self._fetch_dataframe_step(substep, steps_data)
            data = join_query_data(data, sub_data)
        return data

    def _multiple_steps_reduce(self, step, vars, steps_data):
        if step.reduce != 'union':
            raise ErLogicError(f'Unknown MultipleSteps type: {step.reduce}')

        data = ResultSet()

        # mark vars
        steps = []
        for substep in step.steps:
            if isinstance(substep, FetchDataframeStep) is False:
                raise ErLogicError(f'Wrong step type for MultipleSteps: {step}')
            substep = copy.deepcopy(substep)
            markQueryVar(substep.query.where)
            steps.append(substep)

        for var_group in vars:
            steps2 = copy.deepcopy(steps)
            for name, value in var_group.items():
                for substep in steps2:
                    replaceQueryVar(substep.query.where, value, name)
            sub_data = self._multiple_steps(steps2, steps_data)
            data = join_query_data(data, sub_data)

        return data

    def prepare_query(self, prepare=True):
        if prepare:
            # it is prepared statement call
            steps_data = []
            try:
                for step in self.planner.prepare_steps(self.query):
                    data = self.execute_step(step, steps_data)
                    step.set_result(data)
                    steps_data.append(data)
            except PlanningException as e:
                raise ErLogicError(e)

            statement_info = self.planner.get_statement_info()

            self.columns_list = []
            for col in statement_info['columns']:
                self.columns_list.append(
                    Column(
                        database=col['ds'],
                        table_name=col['table_name'],
                        table_alias=col['table_alias'],
                        name=col['name'],
                        alias=col['alias'],
                        type=col['type']
                    )
                )

            self.parameters = [
                Column(
                    name=col['name'],
                    alias=col['alias'],
                    type=col['type']
                )
                for col in statement_info['parameters']
            ]

    def execute_query(self, params=None):
        if self.fetched_data is not None:
            # no need to execute
            return

        steps_data = []
        process_mark = None
        try:
            steps = list(self.planner.execute_steps(params))
            steps_classes = (x.__class__ for x in steps)
            predict_steps = (ApplyPredictorRowStep, ApplyPredictorStep, ApplyTimeseriesPredictorStep)
            if any(s in predict_steps for s in steps_classes):
                process_mark = create_process_mark('predict')
            for step in steps:
                with profiler.Context(f'step: {step.__class__.__name__}'):
                    data = self.execute_step(step, steps_data)
                step.set_result(data)
                steps_data.append(data)
        except PlanningException as e:
            raise ErLogicError(e)
        except Exception as e:
            raise e
        finally:
            if process_mark is not None:
                delete_process_mark('predict', process_mark)

        # save updated query
        self.query = self.planner.query

        # there was no executing
        if len(steps_data) == 0:
            return

        try:
            if self.outer_query is not None:
                # workaround for subqueries in superset. remove it?
                # +++
                # ???

                result = steps_data[-1]
                df = result.to_df()

                df2 = query_df(df, self.outer_query)

                result2 = ResultSet().from_df(df2, database='', table_name='')

                self.columns_list = result2.columns
                self.fetched_data = result2

            else:
                result = steps_data[-1]
                self.fetched_data = result
        except Exception as e:
            raise SqlApiUnknownError("error in preparing result query step") from e

        try:
            if hasattr(self, 'columns_list') is False:
                # how it becomes False?
                self.columns_list = self.fetched_data.columns

            if self.columns_list is None:
                self.columns_list = self.fetched_data.columns

            for col in self.fetched_data.find_columns('__mindsdb_row_id'):
                self.fetched_data.del_column(col)

        except Exception as e:
            raise SqlApiUnknownError("error in column list step") from e

    def execute_step(self, step, steps_data):
        if type(step) == GetPredictorColumns:
            predictor_name = step.predictor.parts[-1]
            dn = self.datahub.get(self.mindsdb_database_name)
            columns = dn.get_table_columns(predictor_name)

            data = ResultSet()
            for column_name in columns:
                data.add_column(Column(
                    name=column_name,
                    table_name=predictor_name,
                    database=self.mindsdb_database_name
                ))

        elif type(step) == GetTableColumns:
            table = step.table
            dn = self.datahub.get(step.namespace)
            ds_query = Select(from_table=Identifier(table), targets=[Star()], limit=Constant(0))

            data, columns_info = dn.query(ds_query, session=self.session)

            data = ResultSet()
            for column in columns_info:
                data.add_column(Column(
                    name=column['name'],
                    type=column.get('type'),
                    table_name=table,
                    database=self.database
                ))

        elif type(step) == FetchDataframeStep:
            data = self._fetch_dataframe_step(step, steps_data)
        elif type(step) == UnionStep:
            left_result = steps_data[step.left.step_num]
            right_result = steps_data[step.right.step_num]

            # count of columns have to match
            if len(left_result.columns) != len(right_result.columns):
                raise ErSqlWrongArguments(
                    f'UNION columns count mismatch: {len(left_result.columns)} != {len(right_result.columns)} ')

            # types have to match
            # TODO: return checking type later
            # for i, left_col in enumerate(left_result.columns):
            #     right_col = right_result.columns[i]
            #     type1, type2 = left_col.type, right_col.type
            #     if type1 is not None and type2 is not None:
            #         if type1 != type2:
            #             raise ErSqlWrongArguments(f'UNION types mismatch: {type1} != {type2}')

            result = ResultSet()
            for col in left_result.columns:
                result.add_column(col)

            records_hashes = []
            for row in left_result.get_records_raw() + right_result.get_records_raw():
                if step.unique:
                    checksum = hashlib.sha256(str(row).encode()).hexdigest()
                    if checksum in records_hashes:
                        continue
                    records_hashes.append(checksum)
                result.add_record_raw(row)

            data = result

        elif type(step) == MapReduceStep:
            try:
                if step.reduce != 'union':
                    raise ErLogicError(f'Unknown MapReduceStep type: {step.reduce}')

                step_data = steps_data[step.values.step_num]
                vars = []
                for row in step_data.get_records():
                    var_group = {}
                    vars.append(var_group)
                    for name, value in row.items():
                        if name != '__mindsdb_row_id':
                            var_group[name] = value

                data = ResultSet()

                substep = step.step
                if type(substep) == FetchDataframeStep:
                    query = substep.query
                    for var_group in vars:
                        markQueryVar(query.where)
                        for name, value in var_group.items():
                            replaceQueryVar(query.where, value, name)
                        sub_data = self._fetch_dataframe_step(substep, steps_data)
                        if len(data.columns) == 0:
                            data = sub_data
                        else:
                            data.add_records(sub_data.get_records())

                        unmarkQueryVar(query.where)
                elif type(substep) == MultipleSteps:
                    data = self._multiple_steps_reduce(substep, vars, steps_data)
                else:
                    raise ErLogicError(f'Unknown step type: {step.step}')

            except Exception as e:
                raise SqlApiUnknownError(f'error in map reduce step: {e}') from e
        elif type(step) == MultipleSteps:
            if step.reduce != 'union':
                raise ErNotSupportedYet(f"Only MultipleSteps with type = 'union' is supported. Got '{step.type}'")
            data = None
            for substep in step.steps:
                subdata = self.execute_step(substep, steps_data)
                if data is None:
                    data = subdata
                else:
                    data.add_records(subdata.get_records())
        elif type(step) == ApplyPredictorRowStep:

            project_name = step.namespace
            predictor_name = step.predictor.parts[0]
            where_data0 = step.row_dict
            project_datanode = self.datahub.get(project_name)

            # fill params
            where_data = {}
            for key, value in where_data0.items():
                if isinstance(value, Parameter):
                    rs = steps_data[value.value.step_num]
                    if rs.length() == 1:
                        # one value, don't do list
                        value = rs.get_records_raw()[0][0]
                    else:
                        value = [i[0] for i in rs.get_records_raw()]
                where_data[key] = value

            version = None
            if len(step.predictor.parts) > 1 and step.predictor.parts[-1].isdigit():
                version = int(step.predictor.parts[-1])

            predictions = project_datanode.predict(
                model_name=predictor_name,
                data=where_data,
                version=version,
                params=step.params,
            )

            columns_dtypes = dict(predictions.dtypes)
            predictions = predictions.to_dict(orient='records')

            # update predictions with input data
            for row in predictions:
                for k, v in where_data.items():
                    if k not in row:
                        row[k] = v

            table_name = get_preditor_alias(step, self.database)

            result = ResultSet()
            result.is_prediction = True
            if len(predictions) > 0:
                cols = list(predictions[0].keys())
            else:
                cols = project_datanode.get_table_columns(predictor_name)

            for col in cols:
                result.add_column(Column(
                    name=col,
                    table_name=table_name[1],
                    table_alias=table_name[2],
                    database=table_name[0],
                    type=columns_dtypes.get(col)
                ))
            result.add_records(predictions)

            data = result

        elif type(step) in (ApplyPredictorStep, ApplyTimeseriesPredictorStep):
            try:
                # set row_id
                data = steps_data[step.dataframe.step_num]

                params = step.params or {}

                for table in data.get_tables()[:1]:  # add  __mindsdb_row_id only for first table
                    row_id_col = Column(
                        name='__mindsdb_row_id',
                        database=table['database'],
                        table_name=table['table_name'],
                        table_alias=table['table_alias']
                    )

                    values = list(range(self.row_id, self.row_id + data.length()))
                    data.add_column(row_id_col, values)
                    self.row_id += data.length()

                project_name = step.namespace
                predictor_name = step.predictor.parts[0]

                where_data = data.get_records()

                # add constants from where
                row_dict = {}
                if step.row_dict is not None:
                    for k, v in step.row_dict.items():
                        if isinstance(v, Result):
                            prev_result = steps_data[v.step_num]
                            # TODO we await only one value: model.param = (subselect)
                            v = prev_result.get_records_raw()[0][0]
                        row_dict[k] = v

                    for record in where_data:
                        record.update(row_dict)

                predictor_metadata = {}
                for pm in self.predictor_metadata:
                    if pm['name'] == predictor_name and pm['integration_name'].lower() == project_name:
                        predictor_metadata = pm
                        break
                is_timeseries = predictor_metadata['timeseries']
                _mdb_forecast_offset = None
                if is_timeseries:
                    if '> LATEST' in self.query_str:
                        # stream mode -- if > LATEST, forecast starts on inferred next timestamp
                        _mdb_forecast_offset = 1
                    elif '= LATEST' in self.query_str:
                        # override: when = LATEST, forecast starts on last provided timestamp instead of inferred next time
                        _mdb_forecast_offset = 0
                    else:
                        # normal mode -- emit a forecast ($HORIZON data points on each) for each provided timestamp
                        params['force_ts_infer'] = True
                        _mdb_forecast_offset = None
                    for row in where_data:
                        if '__mdb_forecast_offset' not in row:
                            row['__mdb_forecast_offset'] = _mdb_forecast_offset

                # for row in where_data:
                #     for key in row:
                #         if isinstance(row[key], datetime.date):
                #             row[key] = str(row[key])

                table_name = get_preditor_alias(step, self.database)
                result = ResultSet()
                result.is_prediction = True

                project_datanode = self.datahub.get(project_name)
                if len(where_data) == 0:
                    cols = project_datanode.get_table_columns(predictor_name) + ['__mindsdb_row_id']
                    for col in cols:
                        result.add_column(Column(
                            name=col,
                            database=table_name[0],
                            table_name=table_name[1],
                            table_alias=table_name[2]
                        ))
                else:
                    predictor_id = predictor_metadata['id']
                    key = f'{predictor_name}_{predictor_id}_{json_checksum(where_data)}'

                    if self.session.predictor_cache is False:
                        data = None
                    else:
                        predictor_cache = get_cache('predict')
                        data = predictor_cache.get(key)

                    if data is None:
                        version = None
                        if len(step.predictor.parts) > 1 and step.predictor.parts[-1].isdigit():
                            version = int(step.predictor.parts[-1])
                        predictions = project_datanode.predict(
                            model_name=predictor_name,
                            data=where_data,
                            version=version,
                            params=params
                        )
                        data = predictions.to_dict(orient='records')
                        columns_dtypes = dict(predictions.dtypes)

                        if data is not None and isinstance(data, list) and self.session.predictor_cache is not False:
                            predictor_cache.set(key, data)
                    else:
                        columns_dtypes = {}
                    if len(data) > 0:
                        cols = list(data[0].keys())
                        for col in cols:
                            result.add_column(Column(
                                name=col,
                                table_name=table_name[1],
                                table_alias=table_name[2],
                                database=table_name[0],
                                type=columns_dtypes.get(col)
                            ))
                    # apply filter
                    if is_timeseries:
                        data = self.apply_ts_filter(data, where_data, step, predictor_metadata)
                    result.add_records(data)

                data = result

            except Exception as e:
                raise SqlApiUnknownError(f'error in apply predictor step: {e}') from e
        elif type(step) == JoinStep:
            try:
                left_data = steps_data[step.left.step_num]
                right_data = steps_data[step.right.step_num]
                table_a, names_a = left_data.to_df_cols(prefix='A')
                table_b, names_b = right_data.to_df_cols(prefix='B')

                if right_data.is_prediction or left_data.is_prediction:
                    # ignore join condition, use row_id
                    a_row_id = left_data.find_columns('__mindsdb_row_id')[0].get_hash_name(prefix='A')
                    b_row_id = right_data.find_columns('__mindsdb_row_id')[0].get_hash_name(prefix='B')

                    join_condition = f'table_a.{a_row_id} = table_b.{b_row_id}'

                    join_type = step.query.join_type.lower()
                    if join_type == 'join':
                        # join type is not specified. using join to prediction data
                        if left_data.is_prediction:
                            join_type = 'left join'
                        elif right_data.is_prediction:
                            join_type = 'right join'
                else:
                    def adapt_condition(node, **kwargs):
                        if not isinstance(node, Identifier) or len(node.parts) != 2:
                            return

                        table_alias, alias = node.parts
                        cols = left_data.find_columns(alias, table_alias)
                        if len(cols) == 1:
                            col_name = cols[0].get_hash_name(prefix='A')
                            return Identifier(parts=['table_a', col_name])

                        cols = right_data.find_columns(alias, table_alias)
                        if len(cols) == 1:
                            col_name = cols[0].get_hash_name(prefix='B')
                            return Identifier(parts=['table_b', col_name])

                    if step.query.condition is None:
                        raise ErNotSupportedYet('Unable to join table without condition')

                    condition = copy.deepcopy(step.query.condition)
                    query_traversal(condition, adapt_condition)

                    join_condition = SqlalchemyRender('postgres').get_string(condition)
                    join_type = step.query.join_type

                query = f"""
                    SELECT * FROM table_a {join_type} table_b
                    ON {join_condition}
                """
                resp_df, _description = query_df_with_type_infer_fallback(query, {
                    'table_a': table_a,
                    'table_b': table_b
                })

                resp_df = resp_df.replace({np.nan: None})

                names_a.update(names_b)
                data = ResultSet().from_df_cols(resp_df, col_names=names_a)

                for col in data.find_columns('__mindsdb_row_id'):
                    data.del_column(col)

            except Exception as e:
                raise SqlApiUnknownError(f'error in join step: {e}') from e

        elif type(step) == FilterStep:
            # used only in join of two regular tables
            result_set = steps_data[step.dataframe.step_num]

            df, col_names = result_set.to_df_cols()
            col_idx = {}
            for name, col in col_names.items():
                col_idx[col.alias] = name
                col_idx[(col.table_alias, col.alias)] = name

            # analyze condition and change name of columns
            def check_fields(node, is_table=None, **kwargs):
                if is_table:
                    raise ErNotSupportedYet('Subqueries is not supported in WHERE')
                if isinstance(node, Identifier):
                    # only column name
                    col_name = node.parts[-1]

                    if len(node.parts) == 1:
                        key = col_name
                    else:
                        table_name = node.parts[-2]
                        key = (table_name, col_name)

                    if key not in col_idx:
                        raise ErKeyColumnDoesNotExist(f'Table not found for column: {key}')

                    new_name = col_idx[key]
                    return Identifier(parts=[new_name])

            where_query = step.query
            query_traversal(where_query, check_fields)

            query_context_controller.remove_lasts(where_query)
            query = Select(targets=[Star()], from_table=Identifier('df'), where=where_query)

            res = query_df(df, query)

            result_set2 = ResultSet().from_df_cols(res, col_names)

            data = result_set2

        elif type(step) == LimitOffsetStep:
            try:
                step_data = steps_data[step.dataframe.step_num]

                step_data2 = ResultSet()
                for col in step_data.columns:
                    step_data2.add_column(col)

                records = step_data.get_records_raw()

                if isinstance(step.offset, int):
                    records = records[step.offset:]
                if isinstance(step.limit, int):
                    records = records[:step.limit]

                for record in records:
                    step_data2.add_record_raw(record)

                data = step_data2

            except Exception as e:
                raise SqlApiUnknownError(f'error in limit offset step: {e}') from e
        elif type(step) == ProjectStep:
            result_set = steps_data[step.dataframe.step_num]

            df, col_names = result_set.to_df_cols()
            col_idx = {}
            tbl_idx = defaultdict(list)
            for name, col in col_names.items():
                col_idx[col.alias] = name
                col_idx[(col.table_alias, col.alias)] = name
                # add to tables
                tbl_idx[col.table_name].append(name)
                if col.table_name != col.table_alias:
                    tbl_idx[col.table_alias].append(name)

            # analyze condition and change name of columns
            def check_fields(node, is_table=None, **kwargs):
                if is_table:
                    raise ErNotSupportedYet('Subqueries is not supported in WHERE')
                if isinstance(node, Identifier):
                    # only column name
                    col_name = node.parts[-1]
                    if isinstance(col_name, Star):
                        if len(node.parts) == 1:
                            # left as is
                            return
                        else:
                            # replace with all columns from table
                            table_name = node.parts[-2]
                            return [
                                Identifier(parts=[col])
                                for col in tbl_idx.get(table_name, [])
                            ]

                    if len(node.parts) == 1:
                        key = col_name
                    else:
                        table_name = node.parts[-2]
                        key = (table_name, col_name)

                    if key not in col_idx:
                        raise ErKeyColumnDoesNotExist(f'Table not found for column: {key}')

                    new_name = col_idx[key]
                    return Identifier(parts=[new_name], alias=node.alias)

            query = Select(
                targets=step.columns,
                from_table=Identifier('df_table')
            )

            targets0 = query_traversal(query.targets, check_fields)
            targets = []
            for target in targets0:
                if isinstance(target, list):
                    targets.extend(target)
                else:
                    targets.append(target)
            query.targets = targets

            res = query_df(df, query)

            result_set2 = ResultSet().from_df_cols(res, col_names, strict=False)

            data = result_set2

        elif type(step) == GroupByStep:
            # used only in join of two regular tables
            step_data = steps_data[step.dataframe.step_num]

            df = step_data.to_df()

            query = Select(targets=step.targets, from_table='df', group_by=step.columns).to_string()
            res = query_df(df, query)

            # stick all columns to first table
            appropriate_table = step_data.get_tables()[0]

            data = ResultSet()

            data.from_df(res, appropriate_table['database'], appropriate_table['table_name'], appropriate_table['table_alias'])

            # columns are changed
            self.columns_list = data.columns

        elif type(step) == SubSelectStep:
            result = steps_data[step.dataframe.step_num]

            table_name = step.table_name
            if table_name is None:
                table_name = 'df_table'
            else:
                table_name = table_name

            query = step.query
            query.from_table = Identifier('df_table')

            if step.add_absent_cols and isinstance(query, Select):
                query_cols = set()

                def f_all_cols(node, **kwargs):
                    if isinstance(node, Identifier):
                        query_cols.add(node.parts[-1])

                query_traversal(query.where, f_all_cols)

                result_cols = [col.name for col in result.columns]

                for col_name in query_cols:
                    if col_name not in result_cols:
                        result.add_column(Column(col_name))

            df = result.to_df()
            res = query_df(df, query)

            result2 = ResultSet()
            # get database from first column
            database = result.columns[0].database
            result2.from_df(res, database, table_name)

            data = result2

        elif type(step) == SaveToTable or type(step) == InsertToTable:
            is_replace = False
            is_create = False

            if type(step) == SaveToTable:
                is_create = True

                if step.is_replace:
                    is_replace = True

            if step.dataframe is not None:
                data = step.dataframe.result_data
            elif step.query is not None:
                data = ResultSet()
                for col in step.query.columns:
                    data.add_column(Column(col.name))

                for row in step.query.values:
                    record = [v.value for v in row]
                    data.add_record_raw(record)
            else:
                raise ErLogicError(f'Data not found for insert: {step}')

            if len(step.table.parts) > 1:
                integration_name = step.table.parts[0]
                table_name = Identifier(parts=step.table.parts[1:])
            else:
                integration_name = self.database
                table_name = step.table

            dn = self.datahub.get(integration_name)

            if hasattr(dn, 'create_table') is False:
                raise ErNotSupportedYet(f"Creating table in '{integration_name}' is not supporting")

            #  del 'service' columns
            for col in data.find_columns('__mindsdb_row_id'):
                data.del_column(col)
            for col in data.find_columns('__mdb_forecast_offset'):
                data.del_column(col)

            # region del columns filtered at projection step
            if self.columns_list is not None:
                filtered_column_names = [x.name for x in self.columns_list]
                for col in data.columns:
                    if col.name.startswith('predictor.'):
                        continue
                    if col.name in filtered_column_names:
                        continue
                    data.del_column(col)
            # endregion

            # drop double names
            col_names = set()
            for col in data.columns:
                if col.alias in col_names:
                    data.del_column(col)
                else:
                    col_names.add(col.alias)

            dn.create_table(
                table_name=table_name,
                result_set=data,
                is_replace=is_replace,
                is_create=is_create
            )
            data = ResultSet()
        elif type(step) == UpdateToTable:
            data = ResultSet()

            if len(step.table.parts) > 1:
                integration_name = step.table.parts[0]
                table_name_parts = step.table.parts[1:]
            else:
                integration_name = self.database
                table_name_parts = step.table.parts

            dn = self.datahub.get(integration_name)

            result = step.dataframe

            params_map_index = []

            if step.update_command.keys is not None:
                result_data = result.result_data

                where = None
                update_columns = {}

                key_columns = [i.to_string() for i in step.update_command.keys]
                if len(key_columns) == 0:
                    raise ErSqlWrongArguments('No key columns in update statement')
                for col in result_data.columns:
                    name = col.name
                    value = Constant(None)

                    if name in key_columns:
                        # put it to where

                        condition = BinaryOperation(
                            op='=',
                            args=[Identifier(name), value]
                        )
                        if where is None:
                            where = condition
                        else:
                            where = BinaryOperation(
                                op='and',
                                args=[where, condition]
                            )
                    else:
                        # put to update
                        update_columns[name] = value

                    params_map_index.append([name, value])

                if len(update_columns) is None:
                    raise ErSqlWrongArguments(f'No columns for update found in: {result_data.columns}')

                update_query = Update(
                    table=Identifier(parts=table_name_parts),
                    update_columns=update_columns,
                    where=where
                )

            else:
                # make command
                update_query = Update(
                    table=Identifier(parts=table_name_parts),
                    update_columns=step.update_command.update_columns,
                    where=step.update_command.where
                )

                if result is None:
                    # run as is
                    dn.query(query=update_query, session=self.session)
                    return data
                result_data = result.result_data

                # link nodes with parameters for fast replacing with values
                input_table_alias = step.update_command.from_select_alias
                if input_table_alias is None:
                    raise ErSqlWrongArguments('Subselect in update requires alias')

                def prepare_map_index(node, is_table, **kwargs):
                    if isinstance(node, Identifier) and not is_table:
                        # is input table field
                        if node.parts[0] == input_table_alias.parts[0]:
                            node2 = Constant(None)
                            param_name = node.parts[-1]
                            params_map_index.append([param_name, node2])
                            # replace node with constant
                            return node2
                        elif node.parts[0] == table_name_parts[0]:
                            # remove updated table alias
                            node.parts = node.parts[1:]

                # do mapping
                query_traversal(update_query, prepare_map_index)

            # check all params is input data:
            data_header = [col.alias for col in result_data.columns]

            for param_name, _ in params_map_index:
                if param_name not in data_header:
                    raise ErSqlWrongArguments(f'Field {param_name} not found in input data. Input fields: {data_header}')

            # perform update
            for row in result_data.get_records():
                # run update from every row from input data

                # fill params:
                for param_name, param in params_map_index:
                    param.value = row[param_name]

                dn.query(query=update_query, session=self.session)
        elif type(step) == DeleteStep:
            if len(step.table.parts) > 1:
                integration_name = step.table.parts[0]
                table_name_parts = step.table.parts[1:]
            else:
                integration_name = self.database
                table_name_parts = step.table.parts

            dn = self.datahub.get(integration_name)

            # make command
            query = Delete(
                table=Identifier(parts=table_name_parts),
                where=copy.deepcopy(step.where),
            )

            # fill params
            def fill_params(node, **kwargs):
                if isinstance(node, Parameter):
                    rs = steps_data[node.value.step_num]
                    items = [Constant(i[0]) for i in rs.get_records_raw()]
                    return Tuple(items)

            query_traversal(query.where, fill_params)

            dn.query(query=query, session=self.session)

            data = ResultSet()

        elif type(step) == DataStep:
            # create resultset
            df = pd.DataFrame(step.data)
            data = ResultSet().from_df(df, database='', table_name='')

        else:
            raise ErLogicError(F'Unknown planner step: {step}')
        return data

    def apply_ts_filter(self, predictor_data, table_data, step, predictor_metadata):

        if step.output_time_filter is None:
            # no filter, exit
            return predictor_data

            # apply filter
        group_cols = predictor_metadata['group_by_columns']
        order_col = predictor_metadata['order_by_column']

        filter_args = step.output_time_filter.args
        filter_op = step.output_time_filter.op

        # filter field must be order column
        if not (
            isinstance(filter_args[0], Identifier)
            and filter_args[0].parts[-1] == order_col
        ):
            # exit otherwise
            return predictor_data

        def get_date_format(samples):
            # dateinfer reads sql date 2020-04-01 as yyyy-dd-mm. workaround for in
            for date_format, pattern in (
                ('%Y-%m-%d', r'[\d]{4}-[\d]{2}-[\d]{2}'),
                ('%Y-%m-%d %H:%M:%S', r'[\d]{4}-[\d]{2}-[\d]{2} [\d]{2}:[\d]{2}:[\d]{2}'),
                # ('%Y-%m-%d %H:%M:%S%z', r'[\d]{4}-[\d]{2}-[\d]{2} [\d]{2}:[\d]{2}:[\d]{2}\+[\d]{2}:[\d]{2}'),
                # ('%Y', '[\d]{4}')
            ):
                if re.match(pattern, samples[0]):
                    # suggested format
                    for sample in samples:
                        try:
                            dt.datetime.strptime(sample, date_format)
                        except ValueError:
                            date_format = None
                            break
                    if date_format is not None:
                        return date_format

            return dateinfer.infer(samples)

        if self.model_types.get(order_col) in ('float', 'integer'):
            # convert strings to digits
            fnc = {
                'integer': int,
                'float': float
            }[self.model_types[order_col]]

            # convert predictor_data
            if len(predictor_data) > 0:
                if isinstance(predictor_data[0][order_col], str):

                    for row in predictor_data:
                        row[order_col] = fnc(row[order_col])
                elif isinstance(predictor_data[0][order_col], dt.date):
                    # convert to datetime
                    for row in predictor_data:
                        row[order_col] = fnc(row[order_col])

            # convert predictor_data
            if isinstance(table_data[0][order_col], str):

                for row in table_data:
                    row[order_col] = fnc(row[order_col])
            elif isinstance(table_data[0][order_col], dt.date):
                # convert to datetime
                for row in table_data:
                    row[order_col] = fnc(row[order_col])

            # convert args to date
            samples = [
                arg.value
                for arg in filter_args
                if isinstance(arg, Constant) and isinstance(arg.value, str)
            ]
            if len(samples) > 0:

                for arg in filter_args:
                    if isinstance(arg, Constant) and isinstance(arg.value, str):
                        arg.value = fnc(arg.value)

        if self.model_types.get(order_col) in ('date', 'datetime') or isinstance(predictor_data[0][order_col], pd.Timestamp):  # noqa
            # convert strings to date
            # it is making side effect on original data by changing it but let it be

            def _cast_samples(data, order_col):
                if isinstance(data[0][order_col], str):
                    samples = [row[order_col] for row in data]
                    date_format = get_date_format(samples)

                    for row in data:
                        row[order_col] = dt.datetime.strptime(row[order_col], date_format)
                elif isinstance(data[0][order_col], dt.datetime):
                    pass  # check because dt.datetime is instance of dt.date but here we don't need to add HH:MM:SS
                elif isinstance(data[0][order_col], dt.date):
                    # convert to datetime
                    for row in data:
                        row[order_col] = dt.datetime.combine(row[order_col], dt.datetime.min.time())

            # convert predictor_data
            if len(predictor_data) > 0:
                _cast_samples(predictor_data, order_col)

            # convert table data
            _cast_samples(table_data, order_col)

            # convert args to date
            samples = [
                arg.value
                for arg in filter_args
                if isinstance(arg, Constant) and isinstance(arg.value, str)
            ]
            if len(samples) > 0:
                date_format = get_date_format(samples)

                for arg in filter_args:
                    if isinstance(arg, Constant) and isinstance(arg.value, str):
                        arg.value = dt.datetime.strptime(arg.value, date_format)
            # TODO can be dt.date in args?

        # first pass: get max values for Latest in table data
        latest_vals = {}
        if Latest() in filter_args:

            for row in table_data:
                if group_cols is None:
                    key = 0  # the same for any value
                else:
                    key = tuple([str(row[i]) for i in group_cols])
                val = row[order_col]
                if key not in latest_vals or latest_vals[key] < val:
                    latest_vals[key] = val

        # second pass: do filter rows
        data2 = []
        for row in predictor_data:
            val = row[order_col]

            if isinstance(step.output_time_filter, BetweenOperation):
                if val >= filter_args[1].value and val <= filter_args[2].value:
                    data2.append(row)
            elif isinstance(step.output_time_filter, BinaryOperation):
                op_map = {
                    '<': '__lt__',
                    '<=': '__le__',
                    '>': '__gt__',
                    '>=': '__ge__',
                    '=': '__eq__',
                }
                arg = filter_args[1]
                if isinstance(arg, Latest):
                    if group_cols is None:
                        key = 0  # the same for any value
                    else:
                        key = tuple([str(row[i]) for i in group_cols])
                    if key not in latest_vals:
                        # pass this row
                        continue
                    arg = latest_vals[key]
                elif isinstance(arg, Constant):
                    arg = arg.value

                if filter_op not in op_map:
                    # unknown operation, exit immediately
                    return predictor_data

                # check condition
                filter_op2 = op_map[filter_op]
                if getattr(val, filter_op2)(arg):
                    data2.append(row)
            else:
                # unknown operation, add anyway
                data2.append(row)

        return data2



import pandas as pd
from mindsdb_sql.parser.ast import Constant, Identifier, Insert, OrderBy, TableColumn
from mindsdb_sql.parser.dialects.mindsdb import (
    CreateJob,
    CreateMLEngine,
    CreatePredictor,
    FinetunePredictor,
    RetrainPredictor,
)

import mindsdb.api.mongo.functions as helpers
from mindsdb.api.mongo.classes import Responder
from mindsdb.api.mongo.classes.query_sql import run_sql_command
from mindsdb.api.mongo.responders.aggregate import aggregate_to_ast
from mindsdb.api.mongo.responders.find import find_to_ast
from mindsdb.api.mongo.utilities.mongodb_parser import MongodbParser
from mindsdb.integrations.libs.response import HandlerStatusResponse
from mindsdb.utilities import log

logger = log.getLogger(__name__)


class Responce(Responder):
    when = {"insert": helpers.is_true}

    def result(self, query, request_env, mindsdb_env, session):
        try:
            res = self._result(query, request_env, mindsdb_env)
        except Exception as e:
            logger.error(e)
            res = {
                "n": 0,
                "writeErrors": [{"index": 0, "code": 0, "errmsg": str(e)}],
                "ok": 1,
            }
        return res

    def _insert_database(self, query, request_env, mindsdb_env):

        for doc in query["documents"]:
            if "_id" in doc:
                del doc["_id"]
            for field in ("name", "engine", "connection_args"):
                if field not in doc:
                    raise Exception(f"'{field}' must be specified")

            status = HandlerStatusResponse(success=False)
            try:
                handler = mindsdb_env["integration_controller"].create_tmp_handler(
                    handler_type=doc["engine"], connection_data=doc["connection_args"]
                )
                status = handler.check_connection()
            except Exception as e:
                status.error_message = str(e)

            if status.success is False:
                raise Exception(f"Can't connect to db: {status.error_message}")

            integration = mindsdb_env["integration_controller"].get(doc["name"])
            if integration is not None:
                raise Exception(f"Database '{doc['name']}' already exists.")

        for doc in query["documents"]:
            mindsdb_env["integration_controller"].add(
                doc["name"], doc["engine"], doc["connection_args"]
            )

    def _insert_model(self, query, request_env, mindsdb_env):
        predictors_columns = [
            "name",
            "status",
            "accuracy",
            "predict",
            "select_data_query",
            "training_options",
            "connection",
        ]

        if len(query["documents"]) != 1:
            raise Exception("Must be inserted just one predictor at time")

        for doc in query["documents"]:
            if "_id" in doc:
                del doc["_id"]

            action = doc.pop("action", "create").lower()

            bad_columns = [x for x in doc if x not in predictors_columns]
            if len(bad_columns) > 0:
                raise Exception(
                    f"Is no possible insert this columns to 'predictors' collection: {', '.join(bad_columns)}"
                )

            if "name" not in doc:
                raise Exception("Please, specify 'name' field")

            if "predict" not in doc:
                predict = None
                if action == "create":
                    raise Exception("Please, specify 'predict' field")
            else:
                predict = doc["predict"]
                if not isinstance(predict, list):
                    predict = [Identifier(x.strip()) for x in predict.split(",")]

            order_by = None
            group_by = None
            ts_settings = {}

            kwargs = doc.get("training_options", {})

            if "timeseries_settings" in kwargs:
                ts_settings = kwargs.pop("timeseries_settings")

                # mongo shell client sends int as float. need to convert it to int
                for key in ("window", "horizon"):
                    val = ts_settings.get(key)
                    if val is not None:
                        ts_settings[key] = int(val)

                if "order_by" in ts_settings:
                    order_by = ts_settings["order_by"]
                    if not isinstance(order_by, list):
                        order_by = [order_by]

                    order_by = [OrderBy(Identifier(x)) for x in order_by]
                if "group_by" in ts_settings:
                    group_by = [Identifier(x) for x in ts_settings.get("group_by", [])]

            using = dict(kwargs)

            select_data_query = doc.get("select_data_query")
            integration_name = None
            if "connection" in doc:
                integration_name = Identifier(doc["connection"])

            Class = CreatePredictor
            if action == "retrain":
                Class = RetrainPredictor
            elif action == "finetune":
                Class = FinetunePredictor

            create_predictor_ast = Class(
                name=Identifier(f"{request_env['database']}.{doc['name']}"),
                integration_name=integration_name,
                query_str=select_data_query,
                targets=predict,
                order_by=order_by,
                group_by=group_by,
                window=ts_settings.get("window"),
                horizon=ts_settings.get("horizon"),
                using=using,
            )

            run_sql_command(request_env, create_predictor_ast)

    def _insert_job(self, query, request_env, mindsdb_env):
        for doc in query["documents"]:

            query_str = doc["query"]
            # try parse as mongo
            parser = MongodbParser()
            try:
                mql = parser.from_string(query_str)

                method = mql.pipeline[0]["method"]
                if method == "find":
                    args = mql.pipeline[0]["args"]
                    projection = None
                    if len(args) > 1:
                        projection = args[1]
                    query = {
                        "find": mql.collection,
                        "filter": args[0],
                        "projection": projection,
                    }

                    for step in mql.pipeline[1:]:
                        if step["method"] == "limit":
                            query["limit"] = step["args"][0]
                        if step["method"] == "skip":
                            query["skip"] = step["args"][0]
                        if step["method"] == "sort":
                            query["sort"] = step["args"][0]
                    # TODO implement group modifiers
                    ast_query = find_to_ast(
                        query, request_env.get("database", "mindsdb")
                    )

                    # to string
                    query_str = ast_query.to_string()
                elif method == "aggregate":
                    query = {
                        "aggregate": mql.collection,
                        "pipeline": mql.pipeline[0]["args"][0],
                    }
                    ast_query = aggregate_to_ast(
                        query, request_env.get("database", "mindsdb")
                    )
                    query_str = ast_query.to_string()
            except Exception:
                # keep query
                pass

            repeat_str = doc.get("schedule_str").lower().lstrip("every ")

            ast_query = CreateJob(
                name=Identifier(doc["name"]),
                start_str=doc.get("start_at"),
                end_str=doc.get("end_at"),
                query_str=query_str,
                repeat_str=repeat_str,
            )
            run_sql_command(request_env, ast_query)

    def _insert_ml_engine(self, query, request_env, mindsdb_env):
        for doc in query["documents"]:

            ast_query = CreateMLEngine(
                name=Identifier(doc["name"]),
                handler=doc["handler"],
                params=doc.get("params"),
            )

            run_sql_command(request_env, ast_query)

    def _result(self, query, request_env, mindsdb_env):
        table = query["insert"]

        if table == "databases":
            self._insert_database(query, request_env, mindsdb_env)

        elif table in ["predictors", "models"]:
            self._insert_model(query, request_env, mindsdb_env)

        elif table == "jobs":
            self._insert_job(query, request_env, mindsdb_env)

        elif table == "ml_engines":
            self._insert_ml_engine(query, request_env, mindsdb_env)

        else:
            # regular insert

            df = pd.DataFrame(query["documents"])
            if "_id" in df.columns:
                df = df.drop("_id", axis=1)

            data = df.to_dict("split")
            values = []
            for row in data["data"]:
                values.append([Constant(i) for i in row])
            ast_query = Insert(
                table=Identifier(table),
                columns=[TableColumn(c) for c in data["columns"]],
                values=values,
            )

            run_sql_command(request_env, ast_query)

        result = {"n": len(query["documents"]), "ok": 1}

        return result


responder = Responce()



import json
import pickle
import traceback
from concurrent import futures

import grpc
from mindsdb.microservices_grpc.executor import executor_pb2_grpc
from mindsdb.microservices_grpc.executor import executor_pb2

from mindsdb.utilities.context import context as ctx
from mindsdb.api.mysql.mysql_proxy.executor.executor import Executor
from mindsdb.api.mysql.mysql_proxy.controllers.session_controller import (
    SessionController,
)
from mindsdb.utilities import log

logger = log.getLogger(__name__)


class SqlServerStub:
    """This class is just an emulation of Server object,
    used by Executor.
    In 'monolithic' mode of MindsDB work the Executor takes
    some information from the sql server which. Here we emulate
    this object."""

    def __init__(self, **kwargs):
        for arg in kwargs:
            setattr(self, arg, kwargs[arg])


class ExecutorServiceServicer(executor_pb2_grpc.ExecutorServiceServicer):

    def __init__(self):
        logger.info(
            "%s.__init__: ", self.__class__.__name__
        )
        self.executors_cache = {}
        self.sessions_cache = {}

    def run(self, **kwargs):
        host = kwargs.get("host", "127.0.0.1")
        port = kwargs.get("port", 50051)
        addr = f"{host}:{port}"
        server = grpc.server(futures.ThreadPoolExecutor(max_workers=10))
        executor_pb2_grpc.add_ExecutorServiceServicer_to_server(
            ExecutorServiceServicer(), server)
        server.add_insecure_port(addr)
        server.start()
        server.wait_for_termination()

    def _get_executor(self, params: executor_pb2.ExecutorContext):
        # We have to send context between client and server
        # here we load the context json received from the client(mindsdb)
        # to the local context instance in this Flask thread
        ctx.load(json.loads(params.context))
        exec_id = params.id
        if exec_id in self.executors_cache:
            logger.info(
                "%s: executor %s found in cache", self.__class__.__name__, exec_id
            )
            return self.executors_cache[exec_id]
        session_id = params.session_id
        if session_id in self.sessions_cache:
            logger.info(
                "%s: session %s found in cache", self.__class__.__name__, session_id
            )
            session = self.sessions_cache[session_id]
        else:
            logger.info(
                "%s: creating new session. id - %s, params - %s",
                self.__class__.__name__,
                session_id,
                params.session,
            )
            session = SessionController()
            self.sessions_cache[session_id] = session
        session_params = json.loads(params.session)
        session.database = session_params["database"]
        session.username = session_params["username"]
        session.auth = session_params["auth"]
        session.prepared_stmts = session_params["prepared_stmts"]
        session.packet_sequence_number = session_params["packet_sequence_number"]
        sqlserver = SqlServerStub(connection_id=params.connection_id)

        logger.info(
            "%s: session info - id=%s, params=%s",
            self.__class__.__name__,
            session_id,
            session.to_json(),
        )
        logger.info(
            "%s: creating new executor. id - %s, session_id - %s",
            self.__class__.__name__,
            exec_id,
            session_id,
        )
        executor = Executor(session, sqlserver)
        self.executors_cache[exec_id] = executor
        return executor

    @staticmethod
    def _prepare_response(executor):
        columns = pickle.dumps(executor.columns)
        params = pickle.dumps(executor.params)
        data = pickle.dumps(executor.data)
        state_track = executor.state_track
        server_status = executor.server_status
        is_executed = executor.is_executed
        session = json.dumps(executor.session.to_json())

        res = executor_pb2.ExecutorResponse(
            columns=columns,
            params=params,
            data=data,
            state_track=state_track,
            server_status=server_status,
            is_executed=is_executed,
            session=session,
        )
        return res

    def DeleteExecutor(self, request, context):
        # to delete executors
        exec_id = request.id
        session_id = request.session_id
        logger.info(
            "%s: removing executor instance. id - %s", self.__class__.__name__, exec_id
        )
        if exec_id is not None and exec_id in self.executors_cache:
            del self.executors_cache[exec_id]

        if session_id is not None and session_id in self.sessions_cache:
            del self.sessions_cache[session_id]

        return executor_pb2.ExecutorStatusResponse(success=True, error_message="")

    def StatementPrepare(self, request, context):
        result = None
        try:
            params = request.json
            executor = self._get_executor(params)
            sql = params.get("sql")
            executor.stmt_prepare(sql)
            result = self._prepare_response(executor)
        except Exception:
            err_msg = traceback.format_exc()
            result = executor_pb2.ExecutorResponse(error_message=err_msg)
        return result

    def StatementExecute(self, request, context):
        result = None
        try:
            executor = self._get_executor(request.context)
            param_values = json.loads(request.param_values)
            executor.stmt_execute(param_values)
            result = self._prepare_response(executor)
        except Exception:
            err_msg = traceback.format_exc()
            result = executor_pb2.ExecutorResponse(error_message=err_msg)
        return result

    def QueryExecute(self, request, context):
        result = None
        try:
            executor = self._get_executor(request.context)
            sql = request.sql
            executor.query_execute(sql)
            result = self._prepare_response(executor)
            result.error_code = executor.error_code if executor.error_code else 0
            result.error_message = executor.error_message if executor.error_message else ""
        except Exception:
            err_msg = traceback.format_exc()
            result = executor_pb2.ExecutorResponse(error_message=err_msg)
        return result

    def BinaryQueryExecute(self, request, context):
        result = None
        try:
            executor = self._get_executor(request.context)
            sql = pickle.loads(request.sql)
            executor.binary_query_execute(sql)
            result = self._prepare_response(executor)
            result.error_code = executor.error_code if executor.error_code else 0
            result.error_message = executor.error_message if executor.error_message else ""
        except Exception:
            err_msg = traceback.format_exc()
            result = executor_pb2.ExecutorResponse(error_message=err_msg)
        return result

    def ExecuteExternal(self, request, context):
        result = None
        try:
            executor = self._get_executor(request.context)
            sql = request.sql
            executor.execute_external(sql)
            result = self._prepare_response(executor)
        except Exception:
            err_msg = traceback.format_exc()
            result = executor_pb2.ExecutorResponse(error_message=err_msg)
        return result

    def Parse(self, request, context):
        result = None
        try:
            executor = self._get_executor(request.context)
            sql = request.sql
            executor.parse(sql)
            result = self._prepare_response(executor)
        except Exception:
            err_msg = traceback.format_exc()
            result = executor_pb2.ExecutorResponse(error_message=err_msg)
        return result

    def DoExecute(self, request, context):
        result = None
        try:
            executor = self._get_executor(request.context)
            executor.do_execute()
            result = self._prepare_response(executor)
        except Exception:
            err_msg = traceback.format_exc()
            result = executor_pb2.ExecutorResponse(error_message=err_msg)
        return result

    def ChangeDefaultDB(self, request, context):
        result = None
        try:
            executor = self._get_executor(request.context)
            new_db = request.new_db
            executor.change_default_db(new_db)
            result = self._prepare_response(executor)
        except Exception:
            err_msg = traceback.format_exc()
            result = executor_pb2.ExecutorResponse(error_message=err_msg)
        return result


def serve():
    host = "127.0.0.1"
    port = 50052
    addr = f"{host}:{port}"
    server = grpc.server(futures.ThreadPoolExecutor(max_workers=10))
    executor_pb2_grpc.add_ExecutorServiceServicer_to_server(
        ExecutorServiceServicer(), server)
    server.add_insecure_port(addr)
    server.start()
    server.wait_for_termination()


if __name__ == '__main__':
    serve()



import json
import pickle
import traceback
from concurrent import futures

import grpc
from mindsdb.microservices_grpc.db import db_pb2_grpc
from mindsdb.microservices_grpc.db import db_pb2, common_pb2

from mindsdb.interfaces.file.file_controller import FileController
from mindsdb.integrations.libs.handler_helpers import get_handler
from mindsdb.api.mysql.mysql_proxy.libs.constants.response_type import RESPONSE_TYPE
from mindsdb.utilities.context import context as ctx
from mindsdb.utilities import log


logger = log.getLogger(__name__)


class DBServiceServicer(db_pb2_grpc.DBServiceServicer):

    def __init__(self):
        logger.info(
            "%s.__init__: ", self.__class__.__name__
        )

    def _get_handler(self, handler_ctx: db_pb2.HandlerContext):
        ctx.load(json.loads(handler_ctx.context))
        handler_class = get_handler(handler_ctx.handler_type)
        logger.info(
            "%s._get_handler: requested instance of %s handler",
            self.__class__.__name__,
            handler_class,
        )
        handler_kwargs = json.loads(handler_ctx.handler_params)
        # Create an instance of FileController for
        # 'files' type of handler
        if handler_ctx.handler_type == "files":
            handler_kwargs["file_controller"] = FileController()
        return handler_class(**handler_kwargs)

    def CheckConnection(self, request, context):

        result = None
        logger.info(
            "%s.check_connection calling", self.__class__.__name__
        )
        try:
            handler = self._get_handler(request)
            res = handler.check_connection()
            result = common_pb2.StatusResponse(success=res.success, error_message=res.error_message)
        except Exception:
            msg = traceback.format_exc()
            result = common_pb2.StatusResponse(success=False, error_message=msg)
        return result

    def Connect(self, request, context):

        result = None
        logger.info(
            "%s.connect calling", self.__class__.__name__
        )
        try:
            handler = self._get_handler(request)
            handler.connect()
            result = common_pb2.StatusResponse(success=True, error_message="")
        except Exception:
            msg = traceback.format_exc()
            result = common_pb2.StatusResponse(success=False, error_message=msg)
        return result

    def Disconnect(self, request, context):
        result = None
        logger.info(
            "%s.disconnect calling", self.__class__.__name__
        )
        try:
            handler = self._get_handler(request)
            handler.disconnect()
            result = common_pb2.StatusResponse(success=True, error_message="")
        except Exception:
            msg = traceback.format_exc()
            result = common_pb2.StatusResponse(success=False, error_message=msg)
        return result

    def NativeQuery(self, request, context):

        result = None
        query = request.query

        logger.info(
            "%s.native_query: calling 'native_query' with query - %s",
            self.__class__.__name__,
            query,
        )
        try:
            handler = self._get_handler(request.context)
            res = handler.native_query(query)
            data = pickle.dumps(res.data_frame)
            result = common_pb2.Response(
                type=res.resp_type,
                data_frame=data,
                query=res.query,
                error_code=res.error_code,
                error_message=res.error_message
            )

        except Exception:
            msg = traceback.format_exc()
            logger.error("%s.native_query: error - %s", self.__class__.__name__, msg)
            result = common_pb2.Response(
                type=RESPONSE_TYPE.ERROR,
                data_frame=None,
                query=0,
                error_code=1,
                error_message=msg,
            )
        return result

    def BinaryQuery(self, request, context):

        result = None
        try:
            query = pickle.loads(request.query)

            logger.info(
                "%s.query: calling 'query' with query - %s",
                self.__class__.__name__,
                query,
            )
            handler = self._get_handler(request.context)
            res = handler.query(query)
            data = pickle.dumps(res.data_frame)
            result = common_pb2.Response(
                type=res.resp_type,
                data_frame=data,
                query=res.query,
                error_code=res.error_code,
                error_message=res.error_message
            )

        except Exception:
            msg = traceback.format_exc()
            logger.error("%s.query: error - %s", self.__class__.__name__, msg)
            result = common_pb2.Response(
                type=RESPONSE_TYPE.ERROR,
                data_frame=None,
                query=0,
                error_code=1,
                error_message=msg
            )
        return result

    def GetTables(self, request, context):

        logger.info(
            "%s.get_tables: calling",
            self.__class__.__name__,
        )
        result = None
        try:
            handler = self._get_handler(request)
            res = handler.get_tables()
            data = pickle.dumps(res.data_frame)
            result = common_pb2.Response(
                type=res.resp_type,
                data_frame=data,
                query=res.query,
                error_code=res.error_code,
                error_message=res.error_message,
            )

        except Exception:
            msg = traceback.format_exc()
            logger.error("%s.get_tables: error - %s", self.__class__.__name__, msg)
            result = common_pb2.Response(
                type=RESPONSE_TYPE.ERROR,
                data_frame=None,
                query=0,
                error_code=1,
                error_message=msg,
            )
        return result

    def GetColumns(self, request, context):

        logger.info(
            "%s.get_columns: calling for table - %s",
            self.__class__.__name__,
            request.table,
        )
        result = None
        try:
            handler = self._get_handler(request.context)
            res = handler.get_columns(request.table)
            data = pickle.dumps(res.data_frame)
            result = common_pb2.Response(
                type=res.resp_type,
                data_frame=data,
                query=res.query,
                error_code=res.error_code,
                error_message=res.error_message,
            )

        except Exception:
            msg = traceback.format_exc()
            logger.error("%s.get_tables: error - %s", self.__class__.__name__, msg)
            result = common_pb2.Response(
                type=RESPONSE_TYPE.ERROR,
                data_frame=None,
                query=0,
                error_code=1,
                error_message=msg,
            )
        return result

    def run(self, **kwargs):
        host = kwargs.get("host", "127.0.0.1")
        port = kwargs.get("port", 50051)
        addr = f"{host}:{port}"
        server = grpc.server(futures.ThreadPoolExecutor(max_workers=10))
        db_pb2_grpc.add_DBServiceServicer_to_server(
            DBServiceServicer(), server)
        server.add_insecure_port(addr)
        # logger.info("staring rpc server on [::]:50051")
        server.start()
        server.wait_for_termination()


def serve():
    server = grpc.server(futures.ThreadPoolExecutor(max_workers=10))
    db_pb2_grpc.add_DBServiceServicer_to_server(
        DBServiceServicer(), server)
    server.add_insecure_port('[::]:50051')
    logger.info("staring rpc server on [::]:50051")
    server.start()
    server.wait_for_termination()


if __name__ == '__main__':
    serve()



from typing import Union, List
import copy
from collections import defaultdict

from mindsdb_sql.parser.ast import (
    Identifier, Select, BinaryOperation, Last, Constant, Star, ASTNode, NullConstant, OrderBy
)
from mindsdb_sql.planner.utils import query_traversal


class LastQuery:
    """
    Wrapper for AST query.
    Intended to ind, track, update last values in query
    """

    def __init__(self, query: ASTNode):
        self.query_orig = None
        self.query = None

        # check query type
        if not isinstance(query, Select):
            # just skip it
            return

        self.last_idx = defaultdict(list)
        last_tables = self.__find_last_columns(query)
        if last_tables is None:
            return

        self.query = query

        self.last_tables = last_tables

    def __find_last_columns(self, query: ASTNode) -> Union[dict, None]:
        """
          This function:
           - Searches LAST column in the input query
           - Replaces it with constants and memorises link to these constants
           - Link to constants will be used to inject values to query instead of LAST
           - Provide checks:
             - if it is possible to find the table for column
             - if column in select target
           - Generates and returns last_column variable which is dict
                last_columns[table_name] = {
                    'table': <table identifier>,
                    'column': <column name>,
                    'links': [<link to ast node>, ... ],
                    'target_idx': <number of column in select target>
                }
        """

        # index last variables in query
        tables_idx = defaultdict(dict)
        conditions = []

        def _index_query(node, is_table, parent_query, **kwargs):

            parent_query_id = id(parent_query)

            if is_table and isinstance(node, Identifier):
                # memorize table
                tables_idx[parent_query_id][node.parts[-1]] = node
                if node.alias is not None:
                    tables_idx[parent_query_id][node.alias.parts[-1]] = node

            # find last in where
            if isinstance(node, BinaryOperation):
                if isinstance(node.args[0], Identifier) and isinstance(node.args[1], Last):
                    # memorize node
                    conditions.append([parent_query_id, node])

        # find lasts
        query_traversal(query, _index_query)

        if len(conditions) == 0:
            return

        self.query_orig = copy.deepcopy(query)

        for parent_query_id, node in conditions:
            # inject constant
            link = Constant(None)
            node.args[1] = link
            self.last_idx[parent_query_id].append(node)

        # index query targets
        query_id = id(query)
        tables = tables_idx[query_id]
        is_star_in_target = False
        target_idx = {}
        for i, target in enumerate(query.targets):
            if isinstance(target, Star):
                is_star_in_target = True
                continue
            elif not isinstance(target, Identifier):
                continue

            col_name = target.parts[-1]
            if len(target.parts) > 1:
                table_name = target.parts[-2]
                table = tables.get(table_name)
            elif len(tables) == 1:
                table = list(tables.values())[0]
            else:
                continue

            target_idx[(table.parts[-1], col_name)] = i

        # make info about query

        last_columns = {}
        for parent_query_id, nodes in self.last_idx.items():
            for node in nodes:
                col, last = node.args
                tables = tables_idx[parent_query_id]

                uniq_tables = len(set([id(v) for v in tables.values()]))
                if len(col.parts) > 1:

                    table = tables.get(col.parts[-2])
                    if table is None:
                        raise Exception('cant find table')
                elif uniq_tables == 1:
                    table = list(tables.values())[0]
                else:
                    # or just skip it?
                    raise Exception('cant find table')

                col_name = col.parts[-1]

                table_name = table.parts[-1]
                if table_name not in last_columns:
                    # check column in target
                    target_idx = target_idx.get((table_name, col_name))
                    if target_idx is None:
                        if is_star_in_target:
                            # will try to get by name
                            ...
                        else:
                            raise Exception('Last value should be in query target')

                    last_columns[table_name] = {
                        'table': table,
                        'column': col_name,
                        'links': [last],
                        'target_idx': target_idx
                    }

                elif last_columns[table_name]['column'] == col_name:
                    last_columns[table_name]['column'].append(last)
                else:
                    raise Exception('possible to use only one column')

        return last_columns

    def to_string(self) -> str:
        """
            String representation of the query
            Used to identify query in query_context table
        """
        return self.query_orig.to_string()

    def get_last_columns(self) -> List[dict]:
        """
        Return information about LAST columns in query
        :return:
        """
        return [
            {
                'table': info['table'],
                'table_name': table_name,
                'column_name': info['column'],
                'target_idx': info['target_idx'],
            }
            for table_name, info in self.last_tables.items()
        ]

    def apply_values(self, values: dict) -> ASTNode:
        """
        Fills query with new values and return it
        """
        for table_name, info in self.last_tables.items():
            value = values.get(table_name, {}).get(info['column'])
            for last in info['links']:
                last.value = value

        return self.query

    def get_init_queries(self):

        back_up_values = []
        # replace values
        for nodes in self.last_idx.values():
            for node in nodes:
                back_up_values.append([node.op, node.args[1]])
                node.op = 'is not'
                node.args[1] = NullConstant()

        query2 = copy.deepcopy(self.query)

        # return values
        for nodes in self.last_idx.values():
            for node in nodes:
                op, arg1 = back_up_values.pop(0)
                node.op = op
                node.args[1] = arg1

        for info in self.get_last_columns():
            col = Identifier(info['column_name'])
            query2.targets = [col]
            query2.order_by = [
                OrderBy(col, direction='DESC')
            ]
            query2.limit = Constant(1)
            yield query2, info



from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional

DESCRIPTOR: _descriptor.FileDescriptor

class Response(_message.Message):
    __slots__ = ["data_frame", "error_code", "error_message", "query", "type"]
    DATA_FRAME_FIELD_NUMBER: _ClassVar[int]
    ERROR_CODE_FIELD_NUMBER: _ClassVar[int]
    ERROR_MESSAGE_FIELD_NUMBER: _ClassVar[int]
    QUERY_FIELD_NUMBER: _ClassVar[int]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    data_frame: bytes
    error_code: int
    error_message: str
    query: int
    type: str
    def __init__(self, type: _Optional[str] = ..., data_frame: _Optional[bytes] = ..., query: _Optional[int] = ..., error_code: _Optional[int] = ..., error_message: _Optional[str] = ...) -> None: ...

class StatusResponse(_message.Message):
    __slots__ = ["error_message", "success"]
    ERROR_MESSAGE_FIELD_NUMBER: _ClassVar[int]
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    error_message: str
    success: bool
    def __init__(self, success: bool = ..., error_message: _Optional[str] = ...) -> None: ...



import json
import pandas as pd

from mindsdb_sql.parser import ast

from mindsdb.integrations.libs.api_handler import APITable
from mindsdb.integrations.utilities.sql_utils import extract_comparison_conditions
from mindsdb.utilities import log
from mindsdb.integrations.libs.response import HandlerResponse as Response


logger = log.getLogger(__name__)


class NotionDatabaseTable(APITable):
    def select(self, query: ast.Select) -> Response:
        conditions = extract_comparison_conditions(query.where)

        params = {}
        filters = []
        for op, arg1, arg2 in conditions:
            if op == "or":
                raise NotImplementedError("OR is not supported")

            if arg1 == "database_id":
                if op == "=":
                    params[arg1] = arg2
                else:
                    NotImplementedError(f"Unknown op: {op}")

            else:
                filters.append([op, arg1, arg2])

        # fetch a particular database with the given id
        # additionally filter the results
        result = self.handler.call_notion_api(
            method_name="databases.query", params=params, filters=filters
        )

        # filter targets
        columns = []
        for target in query.targets:
            if isinstance(target, ast.Star):
                columns = []
                break
            elif isinstance(target, ast.Identifier):
                columns.append(target.parts[-1])
            else:
                raise NotImplementedError

        if len(columns) == 0:
            columns = self.get_columns()

        # columns to lower case
        columns = [name.lower() for name in columns]

        if len(result) == 0:
            result = pd.DataFrame([], columns=columns)
        else:
            # add absent columns
            for col in set(columns) & set(result.columns) ^ set(columns):
                result[col] = None

            # filter by columns
            result = result[columns]
        return result

    def get_columns(self):
        return [
            "id",
            "created_time",
            "last_edited_time",
            "created_by",
            "last_edited_by",
            "cover",
            "icon",
            "parent",
            "archived",
            "properties",
            "url",
            "public_url",
        ]

    def insert(self, query: ast.Insert):
        columns = [col.name for col in query.columns]

        insert_params = ("api_token",)
        for p in insert_params:
            if p not in self.handler.connection_args:
                raise Exception(
                    f"To insert data into Notion, you need to provide the following parameters when connecting it to MindsDB: {insert_params}"
                )  # noqa

        for row in query.values:
            params = dict(zip(columns, row))

            # parent and properties as required params for creating a database
            params["parent"] = json.loads(params["parent"])
            params["properties"] = json.loads(params["properties"])
            params["title"] = json.loads(params.get("title", "{}"))

            self.handler.call_notion_api("databases.create", params)


class NotionPagesTable(APITable):
    def select(self, query: ast.Select) -> Response:
        conditions = extract_comparison_conditions(query.where)

        params = {}
        filters = []
        for op, arg1, arg2 in conditions:
            if op == "or":
                raise NotImplementedError("OR is not supported")

            if arg1 == "page_id":
                if op == "=":
                    params[arg1] = arg2
                else:
                    raise NotImplementedError

            else:
                filters.append([op, arg1, arg2])

        if "query" not in params:
            # search not works without query, use 'mindsdb'
            params["query"] = "mindsdb"

        # fetch a particular page with the given id
        result = self.handler.call_notion_api(
            method_name="pages.retrieve", params=params, filters=filters
        )

        # filter targets
        columns = []
        for target in query.targets:
            if isinstance(target, ast.Star):
                columns = []
                break
            elif isinstance(target, ast.Identifier):
                columns.append(target.parts[-1])
            else:
                raise NotImplementedError

        if len(columns) == 0:
            columns = self.get_columns()

        # columns to lower case
        columns = [name.lower() for name in columns]

        if len(result) == 0:
            result = pd.DataFrame([], columns=columns)
        else:
            # add absent columns
            for col in set(columns) & set(result.columns) ^ set(columns):
                result[col] = None

            # filter by columns
            result = result[columns]
        return result

    def get_columns(self):
        return [
            "id",
            "object",
            "created_time",
            "last_edited_time",
            "created_by",
            "last_edited_by",
            "cover",
            "icon",
            "parent",
            "archived",
            "properties",
            "url",
            "public_url",
        ]

    def insert(self, query: ast.Insert):
        columns = [col.name for col in query.columns]

        insert_params = ("api_token",)
        for p in insert_params:
            if p not in self.handler.connection_args:
                raise Exception(
                    f"To insert data into Notion, you need to provide the following parameters when connecting it to MindsDB: {insert_params}"
                )  # noqa

        for row in query.values:
            params = dict(zip(columns, row))

            # title and database_id as required params for creating the page
            # optionally provide the text to populate the page
            title = params["title"]
            text = params.get("text", "")

            messages = []

            # the last message
            if text.strip() != "":
                messages.append(text.strip())

            len_messages = len(messages)
            for i, text in enumerate(messages):
                if i < len_messages - 1:
                    text += "..."
                else:
                    text += " "

                params["parent"] = {"database_id": params["database_id"]}
                params["properties"] = {
                    "Name": {
                        "title": [
                            {
                                "text": {
                                    "content": title,
                                },
                            },
                        ],
                    },
                }
                params["children"] = [
                    {
                        "object": "block",
                        "type": "paragraph",
                        "paragraph": {
                            "rich_text": [
                                {
                                    "type": "text",
                                    "text": {
                                        "content": text,
                                    },
                                }
                            ]
                        },
                    }
                ]

                self.handler.call_notion_api("pages.create", params)


class NotionBlocksTable(APITable):
    def select(self, query: ast.Select) -> Response:
        conditions = extract_comparison_conditions(query.where)

        params = {}
        filters = []
        for op, arg1, arg2 in conditions:
            if op == "or":
                raise NotImplementedError("OR is not supported")

            if arg1 == "block_id":
                if op == "=":
                    params[arg1] = arg2
                else:
                    NotImplementedError(f"Unknown op: {op}")

            else:
                filters.append([op, arg1, arg2])

        # fetch a particular block with the given id
        result = self.handler.call_notion_api(
            method_name="blocks.retrieve", params=params, filters=filters
        )

        # filter targets
        columns = []
        for target in query.targets:
            if isinstance(target, ast.Star):
                columns = []
                break
            elif isinstance(target, ast.Identifier):
                columns.append(target.parts[-1])
            else:
                raise NotImplementedError

        if len(columns) == 0:
            columns = self.get_columns()

        # columns to lower case
        columns = [name.lower() for name in columns]

        if len(result) == 0:
            result = pd.DataFrame([], columns=columns)
        else:
            # add absent columns
            for col in set(columns) & set(result.columns) ^ set(columns):
                result[col] = None

            # filter by columns
            result = result[columns]
        return result

    def get_columns(self):
        # most of the columns will remain NULL as a block can be of a single type
        return [
            "object",
            "id",
            "parent",
            "has_children",
            "created_time",
            "last_edited_time",
            "created_by",
            "last_edited_by",
            "archived",
            "type",
            "bookmark",
            "breadcrumb",
            "bulleted_list_item",
            "callout",
            "child_database",
            "child_page",
            "column",
            "column_list",
            "divider",
            "embed",
            "equation",
            "file",
            "heading_1",
            "heading_2",
            "heading_3",
            "image",
            "link_preview",
            "link_to_page",
            "numbered_list_item",
            "paragraph",
            "pdf",
            "quote",
            "synced_block",
            "table",
            "table_of_contents",
            "table_row",
            "template",
            "to_do",
            "toggle",
            "unsupported",
            "video",
        ]

    def insert(self, query: ast.Insert):
        columns = [col.name for col in query.columns]

        insert_params = ("api_token",)
        for p in insert_params:
            if p not in self.handler.connection_args:
                raise Exception(
                    f"To insert data into Notion, you need to provide the following parameters when connecting it to MindsDB: {insert_params}"
                )  # noqa

        for row in query.values:
            params = dict(zip(columns, row))

            # block_id and children as required params for appending to a block
            params["block_id"] = params["block_id"]
            params["children"] = json.loads(params["children"])
            params["after"] = params.get("after", "")

            self.handler.call_notion_api("blocks.children.append", params)


class NotionCommentsTable(APITable):
    def select(self, query: ast.Select) -> Response:
        conditions = extract_comparison_conditions(query.where)

        params = {}
        filters = []
        for op, arg1, arg2 in conditions:
            if op == "or":
                raise NotImplementedError("OR is not supported")

            if arg1 == "block_id":
                if op == "=":
                    params[arg1] = arg2
                else:
                    NotImplementedError(f"Unknown op: {op}")

            else:
                filters.append([op, arg1, arg2])

        # list all the unresolved comments for a given block id
        result = self.handler.call_notion_api(
            method_name="comments.list", params=params, filters=filters
        )

        # filter targets
        columns = []
        for target in query.targets:
            if isinstance(target, ast.Star):
                columns = []
                break
            elif isinstance(target, ast.Identifier):
                columns.append(target.parts[-1])
            else:
                raise NotImplementedError

        if len(columns) == 0:
            columns = self.get_columns()

        # columns to lower case
        columns = [name.lower() for name in columns]

        if len(result) == 0:
            result = pd.DataFrame([], columns=columns)
        else:
            # add absent columns
            for col in set(columns) & set(result.columns) ^ set(columns):
                result[col] = None

            # filter by columns
            result = result[columns]
        return result

    def get_columns(self):
        return [
            "id",
            "object",
            "parent",
            "discussion_id",
            "created_time",
            "last_edited_time",
            "created_by",
            "rich_text",
        ]




from contextvars import ContextVar
from typing import Any
from copy import deepcopy


class Context:
    ''' Thread independent storage
    '''
    __slots__ = ('_storage',)

    def __init__(self, storage) -> None:
        object.__setattr__(self, '_storage', storage)
        self.set_default()

    def set_default(self) -> None:
        self._storage.set({
            'company_id': None,
            'user_class': 0,
            'profiling': {
                'level': 0,
                'enabled': False,
                'pointer': None,
                'tree': None
            }
        })

    def __getattr__(self, name: str) -> Any:
        storage = self._storage.get({})
        if name not in storage:
            raise AttributeError(name)
        return storage[name]

    def __setattr__(self, name: str, value: Any) -> None:
        storage = deepcopy(self._storage.get({}))
        storage[name] = value
        self._storage.set(storage)

    def __delattr__(self, name: str) -> None:
        storage = deepcopy(self._storage.get({}))
        if name not in storage:
            raise AttributeError(name)
        del storage['name']
        self._storage.set(storage)

    def dump(self) -> dict:
        storage = deepcopy(self._storage.get({}))
        return storage

    def load(self, storage: dict) -> None:
        self._storage.set(storage)


_context_var = ContextVar('mindsdb.context')
context = Context(_context_var)



import os
import json

import psycopg

from mindsdb.utilities import log

logger = log.getLogger(__name__)


MINDSDB_PROFILING_ENABLED = os.environ.get("MINDSDB_PROFILING_ENABLED") in ("1", "true")
MINDSDB_PROFILING_DB_HOST = os.environ.get("MINDSDB_PROFILING_DB_HOST")
MINDSDB_PROFILING_DB_USER = os.environ.get("MINDSDB_PROFILING_DB_USER")
MINDSDB_PROFILING_DB_PASSWORD = os.environ.get("MINDSDB_PROFILING_DB_PASSWORD")


def set_level(node, level, internal_id):
    internal_id["id"] += 1
    node["level"] = level
    node["value"] = node["stop_at"] - node["start_at"]
    node["value_thread"] = node["stop_at_thread"] - node["start_at_thread"]
    node["value_process"] = node["stop_at_process"] - node["start_at_process"]
    node["internal_id"] = internal_id["id"]

    accum = 0
    for child_node in node["children"]:
        set_level(child_node, level + 1, internal_id)
        accum += child_node["value"]
    node["self"] = node["value"] - accum


def send_profiling_results(profiling_data: dict):
    if MINDSDB_PROFILING_ENABLED is False:
        return

    profiling = profiling_data
    set_level(profiling["tree"], 0, {"id": 0})

    time_start_at = profiling["tree"]["time_start_at"]
    del profiling["tree"]["time_start_at"]

    try:
        connection = psycopg.connect(
            host=MINDSDB_PROFILING_DB_HOST,
            port=5432,
            user=MINDSDB_PROFILING_DB_USER,
            password=MINDSDB_PROFILING_DB_PASSWORD,
            dbname="postgres",
            connect_timeout=5
        )
    except Exception:
        logger.error('cant get acceess to profiling database')
        return
    cur = connection.cursor()
    cur.execute("""
        insert into profiling
            (data, query, time, hostname, environment, api, total_time, company_id, instance_id)
        values
            (%s, %s, %s, %s, %s, %s, %s, %s, %s)
    """, (
        json.dumps(profiling["tree"]),
        profiling.get("query", "?"),
        time_start_at,
        profiling["hostname"],
        profiling.get("environment", "?"),
        profiling.get("api", "?"),
        profiling["tree"]["value"],
        profiling["company_id"],
        profiling["instance_id"]
    ))

    connection.commit()
    cur.close()
    connection.close()



from typing import Optional
from collections import OrderedDict

import pandas as pd
import redshift_connector

from mindsdb_sql import parse_sql
from mindsdb_sql.render.sqlalchemy_render import SqlalchemyRender
from redshift_sqlalchemy import dialect
from mindsdb.integrations.libs.base import DatabaseHandler

from mindsdb_sql.parser.ast.base import ASTNode

from mindsdb.utilities import log
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)
from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE

logger = log.getLogger(__name__)

class RedshiftHandler(DatabaseHandler):
    """
    This handler handles connection and execution of the Redshift statements.
    """

    name = 'redshift'

    def __init__(self, name: str, connection_data: Optional[dict], **kwargs):
        """
        Initialize the handler.
        Args:
            name (str): name of particular handler instance
            connection_data (dict): parameters for connecting to the database
            **kwargs: arbitrary keyword arguments.
        """
        super().__init__(name)
        self.parser = parse_sql
        self.dialect = 'redshift'
        self.connection_data = connection_data
        self.kwargs = kwargs

        self.connection = None
        self.is_connected = False

    def __del__(self):
        if self.is_connected is True:
            self.disconnect()

    def connect(self) -> StatusResponse:
        """
        Set up the connection required by the handler.
        Returns:
            HandlerStatusResponse
        """

        if self.is_connected is True:
            return self.connection

        self.connection = redshift_connector.connect(
            host=self.connection_data['host'],
            port=self.connection_data['port'],
            database=self.connection_data['database'],
            user=self.connection_data['user'],
            password=self.connection_data['password']
        )
        self.is_connected = True

        return self.connection

    def disconnect(self):
        """
        Close any existing connections.
        """

        if self.is_connected is False:
            return

        self.connection.close()
        self.is_connected = False
        return self.is_connected

    def check_connection(self) -> StatusResponse:
        """
        Check connection to the handler.
        Returns:
            HandlerStatusResponse
        """

        response = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            self.connect()
            response.success = True
        except Exception as e:
            logger.error(f'Error connecting to Redshift {self.connection_data["database"]}, {e}!')
            response.error_message = str(e)
        finally:
            if response.success is True and need_to_close:
                self.disconnect()
            if response.success is False and self.is_connected is True:
                self.is_connected = False

        return response

    def native_query(self, query: str) -> StatusResponse:
        """
        Receive raw query and act upon it somehow.
        Args:
            query (str): query in native format
        Returns:
            HandlerResponse
        """

        need_to_close = self.is_connected is False

        connection = self.connect()
        with connection.cursor() as cursor:
            try:
                cursor.execute(query)
                result = cursor.fetchall()
                if result:
                    response = Response(
                        RESPONSE_TYPE.TABLE,
                        data_frame=pd.DataFrame(
                            result,
                            columns=[x[0] for x in cursor.description]
                        )
                    )

                else:
                    response = Response(RESPONSE_TYPE.OK)
                    connection.commit()
            except Exception as e:
                logger.error(f'Error running query: {query} on {self.connection_data["database"]}!')
                response = Response(
                    RESPONSE_TYPE.ERROR,
                    error_message=str(e)
                )

        if need_to_close is True:
            self.disconnect()

        return response

    def query(self, query: ASTNode) -> StatusResponse:
        """
        Receive query as AST (abstract syntax tree) and act upon it somehow.
        Args:
            query (ASTNode): sql query represented as AST. May be any kind
                of query: SELECT, INTSERT, DELETE, etc
        Returns:
            HandlerResponse
        """
        renderer = SqlalchemyRender(dialect.RedshiftDialect)
        query_str = renderer.get_string(query, with_failback=True)
        return self.native_query(query_str)

    def get_tables(self) -> StatusResponse:
        """
        Return list of entities that will be accessible as tables.
        Returns:
            HandlerResponse
        """

        query = """
            SELECT DISTINCT tablename
            FROM PG_TABLE_DEF
            WHERE schemaname = 'public';;
        """
        result = self.native_query(query)
        df = result.data_frame
        result.data_frame = df.rename(columns={'tablename': 'table_name'})
        return result

    def get_columns(self, table_name: str) -> StatusResponse:
        """
        Returns a list of entity columns.
        Args:
            table_name (str): name of one of tables returned by self.get_tables()
        Returns:
            HandlerResponse
        """

        query = f"""
            SELECT column_name,
               data_type,
               CASE WHEN character_maximum_length IS NOT NULL
                    THEN character_maximum_length
                    ELSE numeric_precision end as max_length,
               is_nullable,
               column_default as default_value
            FROM information_schema.columns
            WHERE table_name = '{table_name}' AND table_schema = 'public';
        """
        result = self.native_query(query)
        return result


connection_args = OrderedDict(
    host={
        'type': ARG_TYPE.STR,
        'description': 'The host name or IP address of the Redshift cluster.',
        'required': True,
        'label': 'Host'
    },
    port={
        'type': ARG_TYPE.INT,
        'description': 'The port to use when connecting with the Redshift cluster.',
        'required': True,
        'label': 'Port'
    },
    database={
        'type': ARG_TYPE.STR,
        'description': 'The database name to use when connecting with the Redshift cluster.',
        'required': True,
        'label': 'Database'
    },
    user={
        'type': ARG_TYPE.STR,
        'description': 'The user name used to authenticate with the Redshift cluster.',
        'required': True,
        'label': 'User'
    },
    password={
        'type': ARG_TYPE.PWD,
        'description': 'The password to authenticate the user with the Redshift cluster.',
        'required': True,
        'label': 'Password'
    }
)

connection_args_example = OrderedDict(
    host='examplecluster.abc123xyz789.us-west-1.redshift.amazonaws.com',
    port='5439',
    database='dev',
    user='awsuser',
    password='my_password'
)


from typing import List

import pandas as pd
from mindsdb_sql.parser import ast

from mindsdb.integrations.libs.api_handler import APITable
from mindsdb.integrations.utilities.sql_utils import extract_comparison_conditions
from mindsdb.utilities import log

from mindsdb_sql.parser import ast

logger = log.getLogger(__name__)


class ConfluencePagesTable(APITable):
    """Confluence Pages Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the Confluence "get_all_pages_from_space" API endpoint
        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query
        Returns
        -------
        pd.DataFrame
            confluence "get_all_pages_from_space" matching the query
        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        conditions = extract_comparison_conditions(query.where)

        if query.limit:
            total_results = query.limit.value
        else:
            total_results = 50

        pages_kwargs = {}
        order_by_conditions = {}

        if query.order_by and len(query.order_by) > 0:
            order_by_conditions["columns"] = []
            order_by_conditions["ascending"] = []

            for an_order in query.order_by:
                if an_order.field.parts[0] != "":
                    next
                if an_order.field.parts[1] in self.get_columns():
                    order_by_conditions["columns"].append(an_order.field.parts[1])

                    if an_order.direction == "ASC":
                        order_by_conditions["ascending"].append(True)
                    else:
                        order_by_conditions["ascending"].append(False)
                else:
                    raise ValueError(
                        f"Order by unknown column {an_order.field.parts[1]}"
                    )

        for a_where in conditions:
            if a_where[1] == "space":
                space_name = a_where[2]
                if a_where[0] != "=":
                    raise ValueError("Unsupported where operation for space")
                pages_kwargs["space"] = space_name
            elif a_where[1] not in ["id", "space"]:
                raise ValueError(f"Unsupported where argument {a_where[1]}")

        confluence_pages_records = self.handler.connect().get_all_pages_from_space(
            space_name, start=0, limit=total_results, expand="body.storage"
        )
        confluence_pages_df = pd.json_normalize(confluence_pages_records)

        for a_where in conditions:
            if a_where[1] == "id":
                id = a_where[2]
                if a_where[0] != "=":
                    raise ValueError("Unsupported where operation for id")
                confluence_pages_df = confluence_pages_df[
                    confluence_pages_df.id == id
                ]
                

        def extract_space(input_string):
            parts = input_string.split('/')
            return parts[-1]

        confluence_pages_df["space"] = confluence_pages_df["_expandable.space"].apply(
            extract_space
        )

        confluence_pages_df.columns = confluence_pages_df.columns.str.replace(
            "body.storage.value", "body"
        )

        confluence_pages_df = confluence_pages_df[self.get_columns()]

        if "space" in pages_kwargs:
            confluence_pages_df = confluence_pages_df[
                confluence_pages_df.space == pages_kwargs["space"]
            ]

        selected_columns = []
        for target in query.targets:
            if isinstance(target, ast.Star):
                selected_columns = self.get_columns()
                break
            elif isinstance(target, ast.Identifier):
                selected_columns.append(target.parts[-1])
            else:
                raise ValueError(f"Unknown query target {type(target)}")

        if len(confluence_pages_df) == 0:
            confluence_pages_df = pd.DataFrame([], columns=selected_columns)
        else:
            confluence_pages_df.columns = self.get_columns()
            for col in set(confluence_pages_df.columns).difference(
                set(selected_columns)
            ):
                confluence_pages_df = confluence_pages_df.drop(col, axis=1)

            if len(order_by_conditions.get("columns", [])) > 0:
                confluence_pages_df = confluence_pages_df.sort_values(
                    by=order_by_conditions["columns"],
                    ascending=order_by_conditions["ascending"],
                )

        return confluence_pages_df

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses
        Returns
        -------
        List[str]
            List of columns
        """
        return [
            "id",
            "type",
            "status",
            "title",
            "space",
            "body",
            "_links.self",
            "_links.webui",
        ]

    def insert(self, query: ast.Insert):
        """Inserts a new page into the Confluence space

        Parameters
        ----------
        query : ast.Insert
            Given SQL INSERT query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        columns = [col.name for col in query.columns]

        for i, val in enumerate(query.values):
            params = dict(zip(columns, val))

            self.handler.connect().create_page(
                space=params["space"],
                title=params["title"],
                body=params["body"],
            )

    def update(self, query: ast.Update):
        """Updates a page in the Confluence space

        Parameters
        ----------
        query : ast.Update
            Given SQL UPDATE query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        params = extract_comparison_conditions(query.where)
        title = query.update_columns["title"].get_string()
        body = query.update_columns["body"].get_string()

        for param in params:
            if param[1] == "id":
                id = param[2]
                self.handler.connect().update_page(
                    page_id=id,
                    title=title,
                    body=body,
                )

    def delete(self, query: ast.Delete):
        """Deletes a page from the Confluence space

        Parameters
        ----------
        query : ast.Delete
            Given SQL DELETE query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        params = extract_comparison_conditions(query.where)
        for param in params:
            if param[1] == "id":
                id = param[2]
                self.handler.connect().remove_page(
                    page_id=id,
                )



class BaseEntityException(Exception):
    """Base exception for entitys errors

        Attributes:
            message (str): error message
            entity_name (str): entity name
    """
    def __init__(self, message: str, entity_name: str = None) -> None:
        self.message = message
        self.entity_name = entity_name or 'unkown'

    def __str__(self) -> str:
        return f'{self.message}: {self.entity_name}'


class EntityExistsError(BaseEntityException):
    """Raise when entity exists, but should not"""
    def __init__(self, message: str = None, entity_name: str = None) -> None:
        if message is None:
            message = 'Entity exists error'
        super().__init__(message, entity_name)


class EntityNotExistsError(BaseEntityException):
    """Raise when entity not exists, but should"""
    def __init__(self, message: str = None, entity_name: str = None) -> None:
        if message is None:
            message = 'Entity does not exists error'
        super().__init__(message, entity_name)



from unittest.mock import patch

import pytest
from bs4 import BeautifulSoup
from fitz.fitz import FileDataError

from mindsdb.integrations.handlers.web_handler import urlcrawl_helpers as C
from mindsdb.integrations.handlers.web_handler.tests import example_data as D


class TestPDFToMarkdownTest:
    @patch("requests.Response")
    def test_pdf_to_markdown(self, mock_response) -> None:
        response = mock_response.return_value
        response.content = D.PDF_CONTENT
        result = C.pdf_to_markdown(response)
        assert "Hello, this is a test!" in result

    @patch("requests.Response")
    def test_broken_pdf_to_markdown(self, mock_response) -> None:
        response = mock_response.return_value
        response.content = D.BROKEN_PDF_CONTENT

        with pytest.raises(FileDataError) as excinfo:
            C.pdf_to_markdown(response)

        assert str(excinfo.value) == "cannot open broken document"


@pytest.mark.parametrize(
    "url, result",
    [
        ("google", False),
        ("google.com", False),
        ("https://google.com", True),
        ("", False),
    ],
)
def test_url_validation(url: str, result: bool) -> None:
    assert C.is_valid(url) == result


@pytest.mark.parametrize(
    "html, markdown",
    [(D.HTML_SAMPLE_1, D.MARKDOWN_SAMPLE_1), (D.HTML_SAMPLE_2, D.MARKDOWN_SAMPLE_2)],
)
def test_get_readable_text_from_soup(html: str, markdown: str) -> None:
    soup = BeautifulSoup(html, "html.parser")
    assert markdown == C.get_readable_text_from_soup(soup)



from collections import OrderedDict
import symbl
from mindsdb.integrations.handlers.symbl_handler.symbl_tables import (
    GetConversationTable,
    GetMessagesTable,
    GetTopicsTable,
    GetQuestionsTable,
    GetAnalyticsTable,
    GetActionItemsTable,
    GetFollowUpsTable
)
from mindsdb.integrations.libs.api_handler import APIHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
)
from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE

from mindsdb.utilities import log
from mindsdb_sql import parse_sql


logger = log.getLogger(__name__)


class SymblHandler(APIHandler):
    """The Symbl handler implementation"""

    def __init__(self, name: str, **kwargs):
        """Initialize the Symbl handler.

        Parameters
        ----------
        name : str
            name of a handler instance
        """
        super().__init__(name)

        connection_data = kwargs.get("connection_data", {})
        self.connection_data = connection_data
        self.credentials = {"app_id": self.connection_data.get("app_id"), "app_secret": self.connection_data.get("app_secret")}
        self.kwargs = kwargs
        self.is_connected = False

        conversation_id_data = GetConversationTable(self)
        self._register_table("get_conversation_id", conversation_id_data)

        messages_data = GetMessagesTable(self)
        self._register_table("get_messages", messages_data)

        topics_data = GetTopicsTable(self)
        self._register_table("get_topics", topics_data)

        question_data = GetQuestionsTable(self)
        self._register_table("get_questions", question_data)

        analytics_data = GetAnalyticsTable(self)
        self._register_table("get_analytics", analytics_data)

        ai_data = GetActionItemsTable(self)
        self._register_table("get_action_items", ai_data)

        follow_up_data = GetFollowUpsTable(self)
        self._register_table("get_follow_ups", follow_up_data)

    def connect(self) -> StatusResponse:
        """Set up the connection required by the handler.

        Returns
        -------
        StatusResponse
            connection object
        """
        resp = StatusResponse(False)
        try:
            symbl.AuthenticationToken.get_access_token(self.credentials)
            resp.success = True
            self.is_connected = True
        except Exception as ex:
            resp.success = False
            resp.error_message = ex
            self.is_connected = False
        return resp

    def check_connection(self) -> StatusResponse:
        """Check connection to the handler.

        Returns
        -------
        StatusResponse
            Status confirmation
        """
        return self.connect()

    def native_query(self, query: str) -> StatusResponse:
        """Receive and process a raw query.

        Parameters
        ----------
        query : str
            query in a native format

        Returns
        -------
        StatusResponse
            Request status
        """
        ast = parse_sql(query, dialect="mindsdb")
        return self.query(ast)


connection_args = OrderedDict(
    app_id={
        "type": ARG_TYPE.PWD,
        "description": "App ID of symbl",
        "required": True,
        "label": "app_id",
    },
    app_secret={
        "type": ARG_TYPE.PWD,
        "description": "App Secret of symbl",
        "required": True,
        "label": "app_secret",
    }
)

connection_args_example = OrderedDict(
    app_id="app_id",
    app_secret="app_secret"
)



import re
import numpy as np


def get_completed_prompts(base_template, df):
    """
        Helper method that produces formatted prompts given a template and data in a Pandas DataFrame.
        It also returns the ID of any empty templates that failed to be filled due to missing data.
    """
    columns = []
    spans = []
    matches = list(re.finditer("{{(.*?)}}", base_template))

    assert len(matches) > 0, 'No placeholders found in the prompt, please provide a valid prompt template.'

    first_span = matches[0].start()
    last_span = matches[-1].end()

    for m in matches:
        columns.append(m[0].replace('{', '').replace('}', ''))
        spans.extend((m.start(), m.end()))

    spans = spans[1:-1]  # omit first and last, they are added separately
    template = [base_template[s:e] for s, e in
                list(zip(spans, spans[1:]))[::2]]  # take every other to skip placeholders  # noqa
    template.insert(0, base_template[0:first_span])  # add prompt start
    template.append(base_template[last_span:])  # add prompt end

    empty_prompt_ids = np.where(df[columns].isna().all(axis=1).values)[0]

    df['__mdb_prompt'] = ''
    for i in range(len(template)):
        atom = template[i]
        if i < len(columns):
            col = df[columns[i]].replace(to_replace=[None], value='')  # add empty quote if data is missing
            df['__mdb_prompt'] = df['__mdb_prompt'].apply(lambda x: x + atom) + col.astype("string")
        else:
            df['__mdb_prompt'] = df['__mdb_prompt'].apply(lambda x: x + atom)
    prompts = list(df['__mdb_prompt'])

    return prompts, empty_prompt_ids



from urllib.parse import urlparse
import socket
import ipaddress


def is_private_url(url: str):
    """
    Raises exception if url is private

    :param url: url to check
    """

    hostname = urlparse(url).hostname
    if not hostname:
        # Unable find hostname in url
        return True
    ip = socket.gethostbyname(hostname)
    return ipaddress.ip_address(ip).is_private


def clear_filename(filename: str) -> str:
    """
    Removes path symbols from filename which could be used for path injection
    :param filename: input filename
    :return: output filename
    """

    if not filename:
        return filename
    badchars = '\\/:*?\"<>|'
    for c in badchars:
        filename = filename.replace(c, '')
    return filename



"""projects

Revision ID: 43c52d23845a
Revises: cada7d2be947
Create Date: 2022-10-14 09:59:44.589745

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.sql import text

import mindsdb.interfaces.storage.db as db


# revision identifiers, used by Alembic.
revision = '43c52d23845a'
down_revision = 'cada7d2be947'
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        'project',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.Column('name', sa.String(), nullable=False),
        sa.Column('company_id', sa.Integer(), nullable=True),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('name', 'company_id', name='unique_project_name_company_id')
    )

    conn = op.get_bind()
    session = sa.orm.Session(bind=conn)

    project_record = db.Project(name='mindsdb')
    session.add(project_record)
    session.commit()

    with op.batch_alter_table('predictor', schema=None) as batch_op:
        batch_op.add_column(sa.Column('project_id', sa.Integer()))
        batch_op.create_foreign_key('fk_project_id', 'project', ['project_id'], ['id'])

    conn.execute(sa.sql.text('''
        update predictor set project_id = :project_id
    '''), {'project_id': project_record.id})

    with op.batch_alter_table('predictor', schema=None) as batch_op:
        batch_op.alter_column(
            'project_id',
            existing_type=sa.INTEGER(),
            nullable=False
        )

    with op.batch_alter_table('view', schema=None) as batch_op:
        batch_op.add_column(sa.Column('project_id', sa.Integer()))
        batch_op.create_foreign_key('fk_project_id', 'project', ['project_id'], ['id'])

    conn.execute(sa.sql.text('''
        update view set project_id = :project_id
    '''), {'project_id': project_record.id})

    with op.batch_alter_table('view', schema=None) as batch_op:
        batch_op.alter_column(
            'project_id',
            existing_type=sa.INTEGER(),
            nullable=False
        )

    views = conn.execute(sa.text('''
        select id, name from view
        where exists (select 1 from predictor where view.name = predictor.name)
    ''')).fetchall()

    for row in views:
        conn.execute(
            text("""
                update view
                set name = :name
                where id = :view_id
            """), {
                'name': f"{row['name']}_view",
                'view_id': row['id']
            }
        )

    session.commit()


def downgrade():
    conn = op.get_bind()
    session = sa.orm.Session(bind=conn)

    view_integration = db.Integration.query.filter_by(name='views').first()
    if view_integration is None:
        views_integration = db.Integration(
            name='views',
            data={},
            engine='views',
            company_id=None
        )
        session.add(views_integration)
    session.commit()

    with op.batch_alter_table('view', schema=None) as batch_op:
        batch_op.drop_constraint('fk_project_id', type_='foreignkey')
        batch_op.drop_column('project_id')

    with op.batch_alter_table('predictor', schema=None) as batch_op:
        batch_op.drop_constraint('fk_project_id', type_='foreignkey')
        batch_op.drop_column('project_id')

    op.drop_table('project')



import datetime as dt
from json import JSONEncoder

from bson import ObjectId


class MongoJSONEncoder(JSONEncoder):

    def default(self, obj):
        if isinstance(obj, dt.datetime):
            return f'ISODate({obj.isoformat()})'
        if isinstance(obj, ObjectId):
            return f'ObjectId({str(obj)})'
        return super(MongoJSONEncoder, self).default(obj)


class MongoQuery:

    def __init__(self, collection, pipline=None):
        self.collection = collection
        self.pipeline = []

        if pipline is None:
            pipline = []
        for step in pipline:
            self.add_step(step)

    def add_step(self, step):
        # step = {
        #     'method': 'sort',
        #     'args': [{c: 3}]
        # }

        if 'method' not in step \
                or 'args' not in step \
                or not isinstance(step['args'], list):
            raise AttributeError(f'Wrong step {step}')

        self.pipeline.append(step)

    def to_string(self):
        return self.__str__()

    def __getattr__(self, item):
        # return callback to save step of pipeline
        def fnc(*args):
            self.pipeline.append({
                'method': item,
                'args': args
            })
        return fnc

    def __str__(self):
        """
        converts call to string

        {
            'collection': 'fish',
            'call': [   // call is sequence of methods
                {
                    'method': 'find',
                    'args': [{a:1}, {b:2}]
                },
                {
                    'method': 'sort',
                    'args': [{c:3}]
                },
            ]
        }

        to:

         "db_test.fish.find({a:1}, {b:2}).sort({c:3})"
        """

        call_str = f'db.{self.collection}'
        for step in self.pipeline:
            args_str = []
            for arg in step['args']:
                args_str.append(MongoJSONEncoder().encode(arg))
            call_str += f'.{step["method"]}({",".join(args_str)})'
        return call_str

    def __repr__(self):
        return f'MongoQuery({self.collection}, {str(self.pipeline)})'



from mindsdb.interfaces.storage import db
from mindsdb.integrations.handlers.langchain_embedding_handler.langchain_embedding_handler import construct_model_from_args
from mindsdb.integrations.handlers.langchain_handler.mindsdb_database_agent import MindsDBSQL
from mindsdb.integrations.handlers.rag_handler.rag import RAGQuestionAnswerer
from mindsdb.integrations.handlers.rag_handler.settings import OpenAIParameters, RAGHandlerParameters

import os
from typing import List

_DEFAULT_TOP_K_SIMILARITY_SEARCH = 5


def _make_text_to_sql_tools(skill: db.Skills, llm, executor) -> List:
    # To prevent dependency on Langchain unless an actual tool uses it.
    try:
        from langchain.agents.agent_toolkits import SQLDatabaseToolkit
    except ImportError:
        raise ImportError('To use the text-to-SQL skill, please install langchain with `pip install langchain`')
    database = skill.params['database']
    tables = skill.params['tables']
    tables_to_include = [f'{database}.{table}' for table in tables]
    db = MindsDBSQL(
        engine=executor,
        metadata=executor.session.integration_controller,
        include_tables=tables_to_include
    )
    return SQLDatabaseToolkit(db=db, llm=llm).get_tools()


def _get_rag_query_function(
        skill: db.Skills,
        openai_api_key: str,
        session_controller):

    def _answer_question(question: str) -> str:
        # Get the knowledge base associated with the skill.
        knowledge_base_name = skill.params['source']
        knowledge_base = session_controller.kb_controller.get(knowledge_base_name, skill.project_id)

        # Get embedding model to use when querying the vector store.
        embedding_model_data = session_controller.model_controller.get_model(knowledge_base.embedding_model.name)
        embedding_model_args = embedding_model_data['problem_definition']['using']
        embedding_model_args['target'] = embedding_model_data['problem_definition']['target']
        embedding_model = construct_model_from_args(embedding_model_args)

        # Get vector store handler.
        vector_store_handler_data = session_controller.integration_controller.get(knowledge_base.vector_database.name)
        vector_store_args = vector_store_handler_data['connection_data']
        vector_store_folder_name = vector_store_args.get('persist_directory', os.getcwd())
        vector_store_handler = session_controller.integration_controller.get_handler(knowledge_base.vector_database.name)
        vector_store_storage_path = vector_store_handler.handler_storage.folder_get(vector_store_folder_name)

        # Use OpenAI for interpreting the vector store search response.
        llm_params_dict = {
            'llm_name': 'openai',
            'openai_api_key': openai_api_key
        }
        qa_params = {
            'llm_type': 'openai',
            'llm_params': OpenAIParameters(**llm_params_dict),
            'embeddings_model': embedding_model,
            'top_k': skill.params.get('top_k', _DEFAULT_TOP_K_SIMILARITY_SEARCH),
            'collection_name': knowledge_base.vector_database_table,
            'vector_store_folder_name': vector_store_folder_name,
            'vector_store_storage_path': vector_store_storage_path,
            'vector_store_name': vector_store_handler_data['engine']
        }

        args = RAGHandlerParameters(**qa_params)
        question_answerer = RAGQuestionAnswerer(args=args)
        # Actually answer the query.
        qa_query_response = question_answerer.query(question)
        return qa_query_response['answer']

    return _answer_question


def _make_knowledge_base_tools(
        skill: db.Skills,
        openai_api_key: str,
        session_controller) -> List:
    # To prevent dependency on Langchain unless an actual tool uses it.
    try:
        from langchain.agents import Tool
    except ImportError:
        raise ImportError('To use the knowledge base skill, please install langchain with `pip install langchain`')
    description = skill.params.get('description', '')
    all_tools = []
    all_tools.append(Tool(
        name='Knowledge Base Retrieval',
        func=_get_rag_query_function(skill, openai_api_key, session_controller),
        description=f'Use this tool to get more context or information to answer a question about {description}. The input should be the exact question the user is asking.'
    ))
    return all_tools


def make_tools_from_skill(
        skill: db.Skills,
        llm,
        openai_api_key: str,
        executor) -> List:
    """Makes Langchain compatible tools from a skill

    Args:
        skill (Skills): Skill to make a tool from
        llm (BaseLanguageModel): LLM to use if the skill requires one
        openai_api_key (str): OpenAI API key to use if the skill requires one
        executor (ExecuteCommands): MindsDB executor to use if the skill requires one

    Returns:
        tools (List[BaseTool]): List of tools for the given skill
    """
    if skill.type == 'text_to_sql':
        return _make_text_to_sql_tools(skill, llm, executor)
    elif skill.type == 'knowledge_base':
        return _make_knowledge_base_tools(skill, openai_api_key, executor.session)
    raise NotImplementedError(f'skill of type {skill.type} is not supported as a tool')



"""straighten

Revision ID: 473e8f239481
Revises: 6a54ba55872e
Create Date: 2022-08-29 11:12:11.307317

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.sql import text

import mindsdb.interfaces.storage.db    # noqa


# revision identifiers, used by Alembic.
revision = '473e8f239481'
down_revision = '6a54ba55872e'
branch_labels = None
depends_on = None


def upgrade():
    conn = op.get_bind()
    session = sa.orm.Session(bind=conn)
    conn.execute(text('''
        delete from file
        where exists (
            select 1 from file as t2
            where file.name = t2.name
                  and file.company_id = t2.company_id
                  and file.id < t2.id
        )
    '''))
    session.commit()

    with op.batch_alter_table('file', schema=None) as batch_op:
        batch_op.create_unique_constraint('unique_file_name_company_id', ['name', 'company_id'])

    conn.execute(text('''
        delete from integration where engine is null
    '''))
    session.commit()

    with op.batch_alter_table('integration', schema=None) as batch_op:
        batch_op.alter_column(
            'engine',
            existing_type=sa.VARCHAR(),
            nullable=False
        )

    with op.batch_alter_table('predictor', schema=None) as batch_op:
        batch_op.drop_constraint('unique_predictor_name_company_id', type_='unique')

    conn.execute(text('''
        delete from semaphor
    '''))
    session.commit()

    try:
        with op.batch_alter_table('semaphor', schema=None) as batch_op:
            batch_op.create_unique_constraint('uniq_const', ['entity_type', 'entity_id'])
    except Exception:
        pass


def downgrade():
    with op.batch_alter_table('semaphor', schema=None) as batch_op:
        batch_op.drop_constraint('uniq_const', type_='unique')

    with op.batch_alter_table('predictor', schema=None) as batch_op:
        batch_op.create_unique_constraint('unique_predictor_name_company_id', ['name', 'company_id'])

    with op.batch_alter_table('integration', schema=None) as batch_op:
        batch_op.alter_column(
            'engine',
            existing_type=sa.VARCHAR(),
            nullable=True
        )

    with op.batch_alter_table('file', schema=None) as batch_op:
        batch_op.drop_constraint('unique_file_name_company_id', type_='unique')



from typing import Optional, Union

import requests
from atlassian import Confluence
from mindsdb_sql import parse_sql

from mindsdb.integrations.handlers.confluence_handler.confluence_table import (
    ConfluencePagesTable,
)
from mindsdb.integrations.libs.api_handler import APIHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
)
from mindsdb_sql import parse_sql
from mindsdb.utilities import log

from atlassian import Confluence
from typing import Optional
import requests

logger = log.getLogger(__name__)

class ConfluenceHandler(APIHandler):
    """Confluence handler implementation"""

    def __init__(self, name=None, **kwargs):
        """Initialize the Confluence handler.
        Parameters
        ----------
        name : str
            name of a handler instance
        """
        super().__init__(name)

        connection_data = kwargs.get("connection_data", {})

        self.parser = parse_sql
        self.dialect = 'confluence'
        self.connection_data = connection_data
        self.kwargs = kwargs
        self.connection = None
        self.is_connected = False

        confluence_pages_data = ConfluencePagesTable(self)
        self._register_table("pages", confluence_pages_data)

    def connect(self):
        """Set up the connection required by the handler.
        Returns
        -------
        StatusResponse
            connection object
        """
        if self.is_connected is True:
            return self.connection
        conf = Confluence(
            url=self.connection_data.get('url'),
            username=self.connection_data.get('username'),
            password=self.connection_data.get('password'),
        )
        self.connection = conf
        self.is_connected = True
        return self.connection

    def check_connection(self) -> StatusResponse:
        """Check connection to the handler.
        Returns
        -------
        StatusResponse
            Status confirmation
        """
        response = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            self.connect()
            response.success = True
        except Exception as e:
            logger.error(f"Error connecting to Confluence API: {e}!")
            response.error_message = e

        self.is_connected = response.success

        return response

    def native_query(self, query: str) -> StatusResponse:
        """Receive and process a raw query.
        Parameters
        ----------
        query : str
            query in a native format
        Returns
        -------
        StatusResponse
            Request status
        """
        ast = parse_sql(query, dialect="mindsdb")
        return self.query(ast)



import unittest
from mindsdb.integrations.handlers.google_search_handler.google_search_handler import GoogleSearchConsoleHandler
from mindsdb.api.mysql.mysql_proxy.libs.constants.response_type import RESPONSE_TYPE


class GoogleSearchConsoleHandlerTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.kwargs = {
            "connection_data": {
                "credentials": "/path/to/credentials.json"
            },
            "file_storage": "/path/to/file_storage"
        }
        cls.handler = GoogleSearchConsoleHandler('test_google_search_console_handler', **cls.kwargs)

    def test_0_check_connection(self):
        assert self.handler.check_connection()

    def test_1_get_tables(self):
        tables = self.handler.get_tables()
        assert tables.type is not RESPONSE_TYPE.ERROR

    def test_2_select_analytics_query(self):
        query = "SELECT clicks FROM my_console.Analytics WHERE siteUrl = " \
                "'https://www.mindsdb.com' AND startDate = '2020-10-01' " \
                "AND endDate = '2020-10-31'"
        result = self.handler.native_query(query)
        assert result.type is RESPONSE_TYPE.TABLE

    def test_3_select_sitemaps_query(self):
        query = "SELECT type FROM my_console.Sitemaps WHERE siteUrl = " \
                "'https://www.mindsdb.com' AND feedpath = " \
                "'https://www.mindsdb.com/sitemap.xml'"
        result = self.handler.native_query(query)
        assert result.type is RESPONSE_TYPE.TABLE

    def test_4_insert_sitemaps_query(self):
        query = "INSERT INTO my_console.Sitemaps VALUES " \
                "('https://www.mindsdb.com/sitemap.xml'," \
                "'https://www.mindsdb')"
        result = self.handler.native_query(query)
        assert result.type is RESPONSE_TYPE.TABLE

    def test_5_delete_sitemaps_query(self):
        query = "DELETE FROM my_console.Sitemaps WHERE " \
                "siteUrl = 'https://www.mindsdb' AND feedpath = " \
                "'https://www.mindsdb.com/sitemap.xml'"
        result = self.handler.native_query(query)
        assert result.type is RESPONSE_TYPE.TABLE


if __name__ == '__main__':
    unittest.main()



from mindsdb.integrations.handlers.intercom_handler.intercom_tables import Articles
from mindsdb.integrations.libs.api_handler import APIHandler
from mindsdb.integrations.libs.response import HandlerStatusResponse as StatusResponse
from mindsdb_sql import parse_sql
import requests
import pandas as pd
from collections import OrderedDict
from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE
import json
from mindsdb.utilities import log

logger = log.getLogger(__name__)


class IntercomHandler(APIHandler):
    def __init__(self, name: str, **kwargs) -> None:
        """initializer method

        Args:
            name (str): handler name
        """
        super().__init__(name)

        self.connection = None
        self.is_connected = False
        self._baseUrl = 'https://api.intercom.io'
        args = kwargs.get('connection_data', {})
        if 'access_token' in args:
            access_token = args['access_token']
        self._headers = {
            "Accept": "application/json",
            "Authorization": f"Bearer {access_token}"
        }
        self._register_table(Articles.name, Articles(self))

    def check_connection(self) -> StatusResponse:
        """checking the connection

        Returns:
            StatusResponse: whether the connection is still up
        """
        response = StatusResponse(False)

        try:
            self.connect()
            response.success = True
        except Exception as e:
            logger.error(f'Error connecting to Intercom API: {e}!')
            response.error_message = e

        self.is_connected = response.success
        return response

    def connect(self) -> StatusResponse:
        """making the connectino object
        """
        if self.is_connected and self.connection:
            return self.connection

        if self._headers:
            try:
                response = requests.get(
                    url=self._baseUrl,
                    headers=self._headers
                )
                if response.status_code == 200:
                    self.connection = response
                    self.is_connected = True
                    return StatusResponse(True)
                else:
                    raise Exception(f"Error connecting to Intercom API: {response.status_code} - {response.text}")
            except requests.RequestException as e:
                raise Exception(f"Request to Intercom API failed: {str(e)}")

        raise Exception("Access token is missing")

    def native_query(self, query: str) -> StatusResponse:
        """Receive and process a raw query.

        Parameters
        ----------
        query : str
            query in a native format

        Returns
        -------
        StatusResponse
            Request status
        """
        ast = parse_sql(query, dialect="mindsdb")
        return self.query(ast)

    def call_intercom_api(self, endpoint: str, method: str = 'GET', params: dict = {}, data=None) -> pd.DataFrame:
        url = f"{self._baseUrl}{endpoint}"
        json_data = json.loads(data) if data else None

        response = requests.request(method.upper(), url, headers=self._headers, params=params, json=json_data)

        if response.status_code == 200:
            data = response.json()
            return pd.DataFrame([data])
        else:
            raise requests.Response.raise_for_status(response)


connection_args = OrderedDict(
    access_token={
        "type": ARG_TYPE.PWD,
        "description": "Intercom access token to use for authentication.",
        "required": True,
        "label": "Access token",
    },
)

connection_args_example = OrderedDict(
    api_key="d25509b171ad79395dc2c51b099ee6d0"
)



import os
import json
from collections import OrderedDict

from google.cloud import bigquery
from google.oauth2 import service_account
from sqlalchemy_bigquery.base import BigQueryDialect
from mindsdb_sql.parser.ast.base import ASTNode
from mindsdb_sql.render.sqlalchemy_render import SqlalchemyRender

from mindsdb.integrations.libs.base import DatabaseHandler
from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)
from mindsdb.utilities import log

logger = log.getLogger(__name__)


class BigQueryHandler(DatabaseHandler):
    """
    This handler handles connection and exectuin of Google BigQuery statements
    """
    name = "bigquery"

    def __init__(self, name, connection_data, **kwargs):
        super().__init__(name)
        self.connection_data = connection_data
        self.client = None
        self.is_connected = False

    def __del__(self):
        if self.is_connected is True:
            self.disconnect()

    def _get_account_keys(self):
        if 'service_account_keys' in self.connection_data:
            if os.path.isfile(self.connection_data['service_account_keys']) is False:
                raise Exception("Service account keys' must be path to the file")
            with open(self.connection_data["service_account_keys"]) as source:
                info = json.load(source)
            return info
        elif 'service_account_json' in self.connection_data:
            info = self.connection_data['service_account_json']
            if not isinstance(info, dict):
                raise Exception("service_account_json has to be dict")
            info['private_key'] = info['private_key'].replace('\\n', '\n')
            return info
        else:
            raise Exception('Connection args have to content ether service_account_json or service_account_keys')

    def connect(self):
        """
        Handles the connection to a BigQuery
        """
        if self.is_connected is True:
            return self.client

        info = self._get_account_keys()
        storage_credentials = service_account.Credentials.from_service_account_info(info)
        client = bigquery.Client(
            project=self.connection_data["project_id"],
            credentials=storage_credentials
        )
        self.is_connected = True
        self.client = client
        return self.client

    def check_connection(self) -> StatusResponse:
        """
        Check the connection of the BigQuery
        :return: success status and error message if error occurs
        """
        response = StatusResponse(False)

        try:
            client = self.connect()
            client.query('SELECT 1;')

            # check dataset exists
            client.get_dataset(self.connection_data['dataset'])

            response.success = True
        except Exception as e:
            logger.error(f'Error connecting to BigQuery {self.connection_data["project_id"]}, {e}!')
            response.error_message = e

        if response.success is False and self.is_connected is True:
            self.is_connected = False

        return response

    def native_query(self, query: str) -> Response:
        """
        Receive SQL query and runs it
        :param query: The SQL query to run in BigQuery
        :return: returns the records from the current recordset
        """
        client = self.connect()
        try:
            job_config = bigquery.QueryJobConfig(default_dataset=f"{self.connection_data['project_id']}.{self.connection_data['dataset']}")
            query = client.query(query, job_config=job_config)
            result = query.to_dataframe()
            if not result.empty:
                response = Response(
                    RESPONSE_TYPE.TABLE,
                    result
                )
            else:
                response = Response(RESPONSE_TYPE.OK)
        except Exception as e:
            logger.error(f'Error running query: {query} on {self.connection_data["project_id"]}!')
            response = Response(
                RESPONSE_TYPE.ERROR,
                error_message=str(e)
            )
        return response

    def query(self, query: ASTNode) -> Response:
        """
        Retrieve the data from the SQL statement with eliminated rows that dont satisfy the WHERE condition
        """
        renderer = SqlalchemyRender(BigQueryDialect)
        query_str = renderer.get_string(query, with_failback=True)
        return self.native_query(query_str)

    def get_tables(self) -> Response:
        """
        Get a list with all of the tabels in BigQuery
        """
        q = f"SELECT table_name, table_type, FROM \
             `{self.connection_data['project_id']}.{self.connection_data['dataset']}.INFORMATION_SCHEMA.TABLES`"
        result = self.native_query(q)
        return result

    def get_columns(self, table_name) -> Response:
        """
        Show details about the table
        """
        q = f"SELECT column_name, data_type, FROM \
            `{self.connection_data['project_id']}.{self.connection_data['dataset']}.INFORMATION_SCHEMA.COLUMNS` WHERE table_name = '{table_name}'"
        result = self.native_query(q)
        return result


connection_args = OrderedDict(
    project_id={
        'type': ARG_TYPE.STR,
        'description': 'The BigQuery project id.'
    },
    dataset={
        'type': ARG_TYPE.STR,
        'description': 'The BigQuery dataset name.'
    },
    service_account_keys={
        'type': ARG_TYPE.PATH,
        'description': 'Full path or URL to the service account JSON file'
    },
    service_account_json={
        'type': ARG_TYPE.DICT,
        'description': 'Content of service account JSON file'
    },
)

connection_args_example = OrderedDict(
    project_id='tough-future-332513',
    service_account_keys='/home/bigq/tough-future-332513.json'
)



from collections import OrderedDict

from mindsdb.integrations.handlers.aqicn_handler.aqicn_tables import (
    AQByUserLocationTable,
    AQByCityTable,
    AQByLatLngTable,
    AQByNetworkStationTable
)
from mindsdb.integrations.handlers.aqicn_handler.aqicn import AQIClient
from mindsdb.integrations.libs.api_handler import APIHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
)
from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE

from mindsdb.utilities import log
from mindsdb_sql import parse_sql


logger = log.getLogger(__name__)


class AQICNHandler(APIHandler):
    """The World Air Quality handler implementation"""

    def __init__(self, name: str, **kwargs):
        """Initialize the aqicn handler.

        Parameters
        ----------
        name : str
            name of a handler instance
        """
        super().__init__(name)

        connection_data = kwargs.get("connection_data", {})
        self.connection_data = connection_data
        self.kwargs = kwargs
        self.aqicn_client = None
        self.is_connected = False

        ai_user_location_data = AQByUserLocationTable(self)
        self._register_table("air_quality_user_location", ai_user_location_data)

        ai_city_data = AQByCityTable(self)
        self._register_table("air_quality_city", ai_city_data)

        ai_lat_lng_data = AQByLatLngTable(self)
        self._register_table("air_quality_lat_lng", ai_lat_lng_data)

        aq_network_station_data = AQByNetworkStationTable(self)
        self._register_table("air_quality_station_by_name", aq_network_station_data)

    def connect(self) -> StatusResponse:
        """Set up the connection required by the handler.

        Returns
        -------
        StatusResponse
            connection object
        """
        resp = StatusResponse(False)
        self.aqicn_client = AQIClient(self.connection_data.get("api_key"))
        content = self.aqicn_client.air_quality_user_location()
        if content["code"] != 200:
            resp.success = False
            resp.error_message = content["content"]["data"]
            self.is_connected = False
            return resp
        self.is_connected = True
        resp.success = True
        return resp

    def check_connection(self) -> StatusResponse:
        """Check connection to the handler.

        Returns
        -------
        StatusResponse
            Status confirmation
        """
        response = self.connect()
        self.is_connected = response.success
        return response

    def native_query(self, query: str) -> StatusResponse:
        """Receive and process a raw query.

        Parameters
        ----------
        query : str
            query in a native format

        Returns
        -------
        StatusResponse
            Request status
        """
        ast = parse_sql(query, dialect="mindsdb")
        return self.query(ast)


connection_args = OrderedDict(
    api_key={
        "type": ARG_TYPE.STR,
        "description": "API key",
        "required": True,
        "label": "api_key",
    },
)

connection_args_example = OrderedDict(
    api_key="api_key",
)



from threading import Thread

import mindsdb.interfaces.storage.db as db
from mindsdb.interfaces.stream.utilities import STOP_THREADS_EVENT
from mindsdb.utilities import log
from mindsdb.utilities.context import context as ctx

logger = log.getLogger(__name__)


class Integration:
    def __init__(self, config, name):
        self.config = config
        self.name = name
        self.mindsdb_database = config["api"]["mysql"]["database"]
        self.company_id = ctx.company_id

    def setup(self):
        raise NotImplementedError

    def _query(self, query, fetch=False):
        raise NotImplementedError


class StreamIntegration(Integration):
    def __init__(self, config, name, control_stream=None):
        Integration.__init__(self, config, name)
        self._streams = []
        self._control_stream = control_stream

    def setup(self):
        Thread(target=StreamIntegration._loop, args=(self,)).start()

    def _loop(self):
        logger.info("INTEGRATION %s: starting", self.name)
        while not STOP_THREADS_EVENT.wait(1.0):
            if self._control_stream is not None:
                # Create or delete streams based on messages from control_stream
                for dct in self._control_stream.read():
                    if "action" not in dct:
                        logger.error(
                            "INTEGRATION %s: no action value found in control record - %s",
                            self.name,
                            dct,
                        )
                    else:
                        if dct["action"] == "create":
                            for k in ["name", "predictor", "stream_in", "stream_out"]:
                                if k not in dct:
                                    # Not all required parameters were provided (i.e. stream will not be created)
                                    # TODO: what's a good way to notify user about this?
                                    logger.error(
                                        "INTEGRATION %s: stream creating error. not enough data in control record - %s",
                                        self.name,
                                        dct,
                                    )
                                    break
                            else:
                                logger.info(
                                    "INTEGRATION %s: creating stream %s",
                                    self.name,
                                    dct["name"],
                                )
                                if (
                                    db.session.query(db.Stream)
                                    .filter_by(
                                        name=dct["name"], company_id=self.company_id
                                    )
                                    .first()
                                    is None
                                ):
                                    stream = db.Stream(
                                        company_id=self.company_id,
                                        name=dct["name"],
                                        integration=self.name,
                                        predictor=dct["predictor"],
                                        stream_in=dct["stream_in"],
                                        stream_out=dct["stream_out"],
                                        anomaly_stream=dct.get("stream_anomaly", None),
                                        learning_params=dct.get(
                                            "learning_params", None
                                        ),
                                        learning_threshold=dct.get(
                                            "learning_threshold", None
                                        ),
                                    )
                                    db.session.add(stream)
                                    db.session.commit()
                                else:
                                    logger.error(
                                        "INTEGRATION %s: stream with this name already exists - %s",
                                        self.name,
                                        dct["name"],
                                    )
                        elif dct["action"] == "delete":
                            for k in ["name"]:
                                if k not in dct:
                                    # Not all required parameters were provided (i.e. stream will not be created)
                                    # TODO: what's a good way to notify user about this?
                                    logger.error(
                                        "INTEGRATION %s: unable to delete stream - stream name is not provided",
                                        self.name,
                                    )
                                    break
                            else:
                                logger.error(
                                    "INTEGRATION %s: deleting stream - %s",
                                    self.name,
                                    dct["name"],
                                )
                                db.session.query(db.Stream).filter_by(
                                    company_id=self.company_id,
                                    integration=self.name,
                                    name=dct["name"],
                                ).delete()
                                db.session.commit()
                        else:
                            # Bad action value
                            logger.error(
                                "INTEGRATION %s: bad action value received - %s",
                                self.name,
                                dct,
                            )

            # it is really required to add 'commit' here
            # to avoid case when sessin.query returns previously
            # deleted object
            # https://stackoverflow.com/questions/10210080/how-to-disable-sqlalchemy-caching
            db.session.commit()
            stream_db_recs = (
                db.session.query(db.Stream)
                .filter_by(company_id=self.company_id, integration=self.name)
                .all()
            )

            # Stop streams that weren't found in DB
            indices_to_delete = []
            for i, s in enumerate(self._streams):
                if s.name not in map(lambda x: x.name, stream_db_recs):
                    logger.info(
                        "INTEGRATION %s: stopping stream - %s", self.name, s.name
                    )
                    indices_to_delete.append(i)
                    self._streams[i].stop_event.set()
            self._streams = [
                s for i, s in enumerate(self._streams) if i not in indices_to_delete
            ]

            # Start new streams found in DB
            for s in stream_db_recs:
                if s.name not in map(lambda x: x.name, self._streams):
                    logger.info(
                        "INTEGRATION %s: starting stream - %s", self.name, s.name
                    )
                    self._streams.append(self._make_stream(s))

        logger.info("INTEGRATION %s: stopping", self.name)
        for s in self._streams:
            s.stop_event.set()

    def _make_stream(self, s: db.Stream):
        raise NotImplementedError

    def _query(self, query, fetch=False):
        pass
